"""
Functions related to neutron corrections for changes to atmospheric
pressure.

Functions in this module:

    pressure_correction_l_coeff
    pressure_correction_beta_coeff
    calc_mean_pressure
    calc_beta_coefficient
"""

import numpy as np
from neptoon.logging import get_logger

core_logger = get_logger()


def calc_pressure_correction_beta_coeff(
    current_pressure: float, reference_pressure: float, beta_coeff: float
):
    """
    Calculate a pressure correction factor accounting for changes in
    atmospheric pressure on neutron counting rates, using the beta
    coefficient method.

    Parameters
    ----------
    current_pressure : float
        Current atmospheric pressure at the site in pascals (Pa).
    reference_pressure : float
        Reference atmospheric pressure, typically a long-term average at
        the site, in pascals (Pa) to keep correction factors around 1
    beta_coeff : float
        Beta coefficient, akin to (1/l_coeff), and used as per Hawdon et
        al., 2014, typically around 0.007

    Returns
    -------
    c_factor
        Correction factor to multiply raw neutron count rates by, e.g.,
        1.04.
    """
    if beta_coeff >= 1:
        message = (
            "The beta_coeff is > 1 which suggests "
            "the incorrect function is being used. "
            "Use pressure_correction_l_coeff() instead"
        )

        core_logger.warning(message)
        raise ValueError(message)
    c_factor = np.exp(beta_coeff * (current_pressure - reference_pressure))
    return c_factor


def calc_mean_pressure(elevation: float):
    """
    Calculate mean atmospheric pressure based on elevation.

    Parameters
    ----------
    elevation : float
        Elevation above sea level in meters (m).

    Returns
    -------
    mean_pressure: float
        Mean atmospheric pressure at the given elevation in millibars (mb).
    """
    mean_pressure = (
        101325 * (1 - 2.25577 * (10**-5) * elevation) ** 5.25588
    ) / 100  # output in mb
    return mean_pressure


def calc_beta_coefficient(mean_pressure, latitude, elevation, cutoff_rigidity):
    """
    Calculate the beta coefficient necessary for the
    pressure_correction_beta_coeff() function. This coefficient
    represents the inverse of the mass attenuation length.

    Parameters
    ----------
    mean_pressure : float
        Mean atmospheric pressure in millibars (hPa).
    latitude : float
        Geographic latitude of the site in degrees.
    elevation : float
        Elevation of the site in meters (m).
    cutoff_rigidity : float
        Cutoff rigidity at the site in MeV.

    Returns
    -------
    beta_coeff: float
        Beta coefficient at site (usually ~0.007)
    """
    rho_rck = 2670
    mean_pressure = calc_mean_pressure(elevation)
    # variables
    z = (
        -0.00000448211 * mean_pressure**3
        + 0.0160234 * mean_pressure**2
        - 27.0977 * mean_pressure
        + 15666.1
    )

    # latitudeitude correction
    g_lat = 978032.7 * (
        1
        + 0.0053024 * (np.sin(np.radians(latitude)) ** 2)
        - 0.0000058 * (np.sin(np.radians(2 * latitude))) ** 2
    )

    # free air correction
    del_free_air = -0.3087691 * z

    # Bouguer correction
    del_boug = rho_rck * z * 0.00004193

    g_corr = (g_lat + del_free_air + del_boug) / 100000

    # final gravity and depth
    g = g_corr / 10
    x = mean_pressure / g

    # --- elevation scaling ---

    # parameters
    n_1 = 0.01231386
    alpha_1 = 0.0554611
    k_1 = 0.6012159
    b0 = 4.74235e-06
    b1 = -9.66624e-07
    b2 = 1.42783e-09
    b3 = -3.70478e-09
    b4 = 1.27739e-09
    b5 = 3.58814e-11
    b6 = -3.146e-15
    b7 = -3.5528e-13
    b8 = -4.29191e-14

    # calculations
    term1 = (
        n_1
        * (1 + np.exp(-alpha_1 * cutoff_rigidity**k_1)) ** -1
        * (x - mean_pressure)
    )
    term2 = (
        0.5
        * (b0 + b1 * cutoff_rigidity + b2 * cutoff_rigidity**2)
        * (x**2 - mean_pressure**2)
    )
    term3 = (
        0.3333
        * (b3 + b4 * cutoff_rigidity + b5 * cutoff_rigidity**2)
        * (x**3 - mean_pressure**3)
    )
    term4 = (
        0.25
        * (b6 + b7 * cutoff_rigidity + b8 * cutoff_rigidity**2)
        * (x**4 - mean_pressure**4)
    )

    beta_ceoff = abs((term1 + term2 + term3 + term4) / (mean_pressure - x))

    return beta_ceoff


def dunai_2020(
    current_pressure: float,
    reference_pressure: float,
    beta_coeff: float,
    inclination: float,
):
    """
    TODO
    !!!Speak with Martin about this method from corny!!!
    """
    pass



import logging
from .log_utils import get_log_path


def configure_logger():
    """
    Configuration steps for the core_logger. This will record errors,
    warnings etc. with time stamps for debugging.

    Returns
    -------
    logging.Logger
        The core logger
    """
    logger = logging.getLogger("global_logger")
    log_file_path = get_log_path()
    log_file_path.parent.mkdir(
        parents=True, exist_ok=True
    )  # Create the directory if it doesn't exist
    logger_handler = logging.FileHandler(log_file_path)
    logger_handler.setLevel(logging.WARNING)
    f_format = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    logger_handler.setFormatter(f_format)
    logger.addHandler(logger_handler)
    return logger


logger = configure_logger()


def get_logger():
    """
    Retrieve the logger after it has been instantiated.

    Returns
    -------
    logging.Logger
        The code logger
    """
    return logger



from typing import List, Optional, Literal, Dict
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pathlib import Path
from datetime import datetime
import yaml

from neptoon.logging import get_logger
from enum import Enum


core_logger = get_logger()


class BaseConfig(BaseModel):
    """
    Base configuration class with flexible field allowance.
    All configuration models inherit from this to maintain consistency.
    """

    model_config = ConfigDict(
        extra="allow",
        frozen=False,
    )


# Site Metadata Validation


class SensorInfo(BaseConfig):
    """General site metadata section."""

    name: str = Field(
        description="The name of the site.",
        examples=["Sheepdrove", "Gatton"],
    )
    country: str = Field(
        description="The country the site is located in.",
        examples=["DEU", "UK", "USA", "KOR"],
    )
    identifier: str = Field(
        description="A unique identier",
        examples=["101", "456"],
    )
    install_date: datetime
    latitude: float
    longitude: float
    elevation: float
    time_zone: int
    site_cutoff_rigidity: float
    avg_lattice_water: Optional[float] = Field(default=None)
    avg_soil_organic_carbon: Optional[float] = Field(default=None)
    avg_dry_soil_bulk_density: Optional[float] = Field(default=None)
    N0: Optional[int] = Field(
        gt=0, description="The N0 calibration term.", default=None
    )
    beta_coefficient: Optional[float] = Field(default=None)
    mean_pressure: Optional[float] = Field(default=None)


# Time Series Validation


class TimeSeriesColumns(BaseConfig):
    """
    Defines the structure for column configurations while allowing
    extensions.
    """

    epithermal_neutron_columns: List[str]
    thermal_neutron_columns: Optional[List[str]] = None
    neutron_count_units: Literal[
        "absolute_count", "counts_per_hour", "counts_per_second"
    ]
    pressure_columns: List[str]
    pressure_units: Optional[str] = None
    pressure_merge_method: Optional[Literal["priority", "average"]] = (
        "priority"
    )
    temperature_columns: List[str]
    temperature_merge_method: Optional[Literal["priority", "average"]] = (
        "priority"
    )
    relative_humidity_columns: List[str]
    relative_humidity_units: Optional[str] = None
    relative_humidity_merge_method: Optional[
        Literal["priority", "average"]
    ] = "priority"
    date_time_columns: List[str]
    date_time_format: str


class Temporal(BaseConfig):
    input_resolution: str = Field(default="1hour")
    output_resolution: str = Field(default=None)
    align_timestamps: bool = Field(default=False)
    alignment_method: str | None = Field(default="time")
    aggregate_method: str | None = Field(default="bagg")
    aggregate_func: str | None = Field(default="mean")
    aggregate_maxna_fraction: float | None = Field(default=0.5)


class TimeSeriesData(BaseConfig):
    path_to_data: Optional[str] = Field(default=None)
    temporal: Optional[Temporal] = None
    key_column_info: Optional[TimeSeriesColumns] = None


# Parser Validation


class ParserKeywords(BaseModel):
    """Configuration for specific parser keywords."""

    strip_left: bool = Field(default=True, description="TODO")
    digit_first: bool = Field(
        default=True,
        description="TODO",
    )


class RawDataParseConfig(BaseModel):
    """Configuration for parsing raw data files."""

    parse_raw_data: bool = Field(
        default=True, description="Whether to parse raw data files"
    )

    data_location: Optional[Path] = Field(
        description="Path to the raw data files or directory"
    )

    column_names: Optional[List[str]] = Field(
        default=None,
        description="A list of the raw column names in the order they appear",
    )

    prefix: Optional[str] = Field(
        default="",
        description="Prefix of file name for file filtering",
    )

    suffix: Optional[str] = Field(
        default="",
        description="Suffix of file name used for file filtering",
    )

    encoding: Optional[str] = Field(
        default="cp850",
        description="File encoding format",
    )

    skip_lines: Optional[int] = Field(
        default=0,
        description="Number of lines to skip at start of file",
    )

    separator: Optional[str] = Field(
        default=",",
        description="Column separator character",
    )

    decimal: Optional[str] = Field(
        default=".",
        description="Decimal point character",
    )

    skip_initial_space: Optional[bool] = Field(
        default=True,
        description="Whether to skip initial whitespace when making dataframe",
    )

    parser_kw: Optional[ParserKeywords] = Field(
        default_factory=ParserKeywords,
        description="Additional parser-specific keywords",
    )

    starts_with: Optional[str] = Field(
        default="",
        description="String that headers must start with when parsing",
    )

    multi_header: Optional[bool] = Field(
        default=False,
        description="Whether to expect multi-line headers",
    )

    strip_names: Optional[bool] = Field(
        default=True,
        description="Whether to strip whitespace from column names",
    )

    remove_prefix: Optional[str] = Field(
        default="//",
        description="Prefix to remove from column names",
    )


# QA Validation


class FlagRange(BaseConfig):
    """Common pattern for min/max range checks."""

    min: Optional[float] = Field(
        default=float("-inf"),
        description="minimum value below which data is removed",
    )
    max: Optional[float] = Field(
        default=float("inf"),
        description="maximum value below which data is removed",
    )

    @model_validator(mode="after")
    def validate_range(cls, values):
        """Validate min is less than max after both fields are set."""
        min_val = values.min if values.min is not None else float("-inf")
        max_val = values.max if values.max is not None else float("inf")

        if min_val > max_val:
            raise ValueError(
                f"min value ({min_val}) must be less than max value ({max_val})"
            )

        return values


class PersistenceCheck(BaseConfig):
    """Configuration for persistence checking."""

    threshold: Optional[float] = None
    window: Optional[int] = None
    min_periods: Optional[int] = None


class SpikeUniLOF(BaseConfig):

    periods_in_calculation: Optional[int] = 20
    threshold: Optional[float] = 1.5
    algorithm: Optional[Literal["ball_tree", "kd_tree", "brute", "auto"]] = (
        Field(default="ball_tree")
    )


class GreaterThanN0(BaseConfig):

    percent_maximum: float = Field(
        default=1.075,
        description="The factor above N0 to flag values",
    )


class BelowN0Factor(BaseConfig):

    percent_minimum: float = Field(
        default=0.3,
        description=(
            "The proportion of N0 value below " "which neutrons are flagged"
        ),
    )


class QAColumnConfig(BaseConfig):
    """
    Base configuration for QA columns.

    Includes all possible QA systems as optional.
    """

    flag_range: Optional[FlagRange] = None
    persistance_check: Optional[PersistenceCheck] = None
    spike_uni_lof: Optional[SpikeUniLOF] = None
    greater_than_N0: Optional[GreaterThanN0] = None
    below_N0_factor: Optional[BelowN0Factor] = None


class QAConfig(BaseConfig):
    """Quality assessment configuration section."""

    air_humidity: Optional[QAColumnConfig] = None
    air_pressure: Optional[QAColumnConfig] = None
    temperature: Optional[QAColumnConfig] = None
    soil_moisture: Optional[QAColumnConfig] = None


class SoilMoistureQA(BaseConfig):
    soil_moisture: Optional[QAColumnConfig] = None


# Calibration Validation


class CalibrationColumnNames(BaseConfig):
    """Column naming configuration for calibration data."""

    date_time: str = Field(default="date_time")
    profile_id: str = Field(default="profile_id")
    sample_depth: str = Field(default="sample_depth")
    radial_distance_from_sensor: str = Field(
        default="radial_distance_from_sensor"
    )
    bulk_density_of_sample: str = Field(default="bulk_density_of_sample")
    gravimetric_soil_moisture: str = Field(
        default="soil_moisture_gravimetric_column"
    )
    soil_organic_carbon: str = Field(default="soil_organic_carbon")
    lattice_water: str = Field(default="lattice_water")


class CalibrationConfig(BaseConfig):
    """Configuration for calibration data."""

    calibrate: bool = Field(default=False)

    data_format: Optional[
        Literal["custom", "cosmoz", "cosmos-usa", "cosmos-uk"]
    ] = Field(default="custom")
    location: Optional[Path] = Field(default="")
    key_column_names: Optional[CalibrationColumnNames] = None


class DataStorageConfig(BaseConfig):
    save_folder: Optional[str] = Field(default=None)
    append_timestamp_to_folder_name: Optional[bool] = Field(default=True)
    append_audit_log_hash_to_folder_name: Optional[bool] = Field(default=False)
    create_report: Optional[bool] = Field(default=False)


class FiguresConfig(BaseConfig):
    create_figures: bool = Field(default=True)
    make_all_figures: bool = Field(default=True)
    custom_list: Optional[List[str]] = Field(
        default=None,
        description="A list of the figures to process",
    )


class SensorConfig(BaseConfig):
    """Top-level configuration."""

    sensor_info: SensorInfo
    time_series_data: Optional[TimeSeriesData] = None
    input_data_qa: Optional[QAConfig] = None
    soil_moisture_qa: Optional[SoilMoistureQA] = None
    raw_data_parse_options: Optional[RawDataParseConfig] = None
    calibration: Optional[CalibrationConfig] = None
    data_storage: Optional[DataStorageConfig] = None
    figures: Optional[FiguresConfig] = None


####################
## Process Config ##
####################


class NeutronQualityAssessment(BaseConfig):
    """Quality assessment configuration for Neutrons"""

    raw_neutrons: QAColumnConfig
    corrected_neutrons: QAColumnConfig


class ReferenceNeutronMonitor(BaseModel):
    """Configuration for reference neutron monitor settings."""

    station: str = Field(
        default="JUNG", description="Station identifier for neutron monitoring"
    )
    resolution: int = Field(
        default=60, description="Time resolution in minutes", ge=1
    )
    nmdb_table: str = Field(
        default="revori", description="NMDB table name to query"
    )


class AirHumidityCorrection(BaseModel):
    """Configuration for air humidity correction parameters."""

    method: Literal["rosolem_2013"] = Field(
        description="Air humidity correction method"
    )
    omega: float = Field(
        default=0.0054,
        description="Omega coefficient for humidity correction",
        gt=0,
    )
    humidity_ref: float = Field(
        default=0, description="Reference humidity value"
    )


class AirPressureCorrection(BaseModel):
    """Configuration for air pressure correction parameters."""

    method: Literal["zreda_2012"] = Field(
        description="Air pressure correction method"
    )
    dunai_inclination: Optional[float] = Field(
        default=None, description="Dunai inclination parameter"
    )


class SoilMoistureEstimation(BaseModel):
    """Configuration for the conversion of neutrons to soil moisture"""

    method: Literal["desilets_etal_2010", "koehli_etal_2021"] = Field(
        description="Soil moisture estimation theory",
        default="desilets_etal_2010",
    )
    koehli_method_form: Literal[
        "Jan23_uranos",
        "Jan23_mcnpfull",
        "Mar12_atmprof",
        "Mar21_mcnp_drf",
        "Mar21_mcnp_ewin",
        "Mar21_uranos_drf",
        "Mar21_uranos_ewin",
        "Mar22_mcnp_drf_Jan",
        "Mar22_mcnp_ewin_gd",
        "Mar22_uranos_drf_gd",
        "Mar22_uranos_ewin_chi2",
        "Mar22_uranos_drf_h200m",
        "Aug08_mcnp_drf",
        "Aug08_mcnp_ewin",
        "Aug12_uranos_drf",
        "Aug12_uranos_ewin",
        "Aug13_uranos_atmprof",
        "Aug13_uranos_atmprof2",
    ] = Field(
        description="Koehli specific method for converting neutrons",
        default="Mar21_uranos_drf",
    )


class IncomingRadiationCorrection(BaseModel):
    """Configuration for incoming radiation correction parameters."""

    method: Literal[
        "zreda_2012",
        "hawdon_2014",
        "mcjannet_desilets_2023",
    ] = Field(description="Incoming radiation correction method")

    reference_neutron_monitor: ReferenceNeutronMonitor = Field(
        default_factory=ReferenceNeutronMonitor,
        description="Reference neutron monitor configuration",
    )


class BiomassCorrection(BaseModel):
    """Configuration for above ground biomass correction parameters."""

    method: Optional[str] = Field(
        default=None, description="Above ground biomass correction method"
    )


class CorrectionSteps(BaseModel):
    """Main configuration for all correction steps."""

    air_humidity: Optional[AirHumidityCorrection] = Field(
        default=None, description="Air humidity correction configuration"
    )
    air_pressure: Optional[AirPressureCorrection] = Field(
        default=None, description="Air pressure correction configuration"
    )
    incoming_radiation: Optional[IncomingRadiationCorrection] = Field(
        default=None, description="Incoming radiation correction configuration"
    )
    above_ground_biomass: Optional[BiomassCorrection] = Field(
        default=None,
        description="Above ground biomass correction configuration",
    )
    soil_moisture_estimation: Optional[SoilMoistureEstimation] = Field(
        default=None, description="Soil Moisture estimation configuration"
    )


class SmoothingAlgorithmSettings(BaseModel):
    """
    Configuration settings for data smoothing algorithms.

    Validates and enforces constraints specific to different smoothing methods:
    - Window size must be positive
    - For Savitzky-Golay:
        - Window size should be odd
        - Polynomial order must be less than window size
    """

    algorithm: Literal["savitsky_golay", "rolling_mean"] = Field(
        default="savitsky_golay", description="Smoothing algorithm to apply"
    )
    window: str = Field(
        default="12h", description="Temporal size of window for smoothing"
    )
    poly_order: Optional[int] = Field(
        default=4,
        description="Polynomial order for Savitzky_Golay filter",
        ge=0,
    )

    @model_validator(mode="after")
    def validate_poly_order(self) -> "SmoothingAlgorithmSettings":
        """Validate polynomial order relative to window size."""
        if self.algorithm == "savitsky_golay":
            if self.poly_order >= self.window:
                raise ValueError(
                    "Polynomial order must be less than window size "
                    f"(got order={self.poly_order}, window={self.window})"
                )
        return self


class DataSmoothingConfig(BaseModel):
    """
    Main configuration for data smoothing operations.

    Controls which data series should be smoothed and defines the
    smoothing parameters to be applied.
    """

    smooth_corrected_neutrons: bool = Field(
        default=True, description="Apply smoothing to corrected neutron counts"
    )
    smooth_soil_moisture: bool = Field(
        default=False,
        description="Apply smoothing to calculated soil moisture values",
    )
    settings: SmoothingAlgorithmSettings = Field(
        default_factory=SmoothingAlgorithmSettings,
        description="Smoothing algorithm configuration",
    )


class ProcessConfig(BaseConfig):

    neutron_quality_assessment: NeutronQualityAssessment
    correction_steps: CorrectionSteps
    data_smoothing: DataSmoothingConfig


class ConfigType(Enum):
    SENSOR = "sensor"
    PROCESS = "process"


class ConfigurationManager:
    """Manages loading and access of nested configurations."""

    def __init__(self):
        self._configs: Dict[str, BaseConfig] = {}

    def _get_working_directory(
        self,
        config_dict: dict,
        config_path: Path,
    ):
        if "working_directory" in config_dict and isinstance(
            config_dict["working_directory"], str
        ):
            parent_path = Path(config_dict["working_directory"])
        else:
            parent_path = config_path.parent
        return parent_path

    def _resolve_paths(
        self,
        config_dict: dict,
        config_path: Path,
        parent_path: Optional[Path] = None,
    ):
        """
        Resolves the paths in the YAML file so that any relative paths
        are relative to the config file itself. The given paths are
        resolved to Path objects.

        Parameters
        ----------
        config_dict : dict
            The loaded YAML file as a dict
        config_path : Path
            The Path of the config YAML

        Returns
        -------
        dict
            Dictionary with paths as Path objects.
        """
        if parent_path is None:
            parent_path = self._get_working_directory(
                config_dict=config_dict, config_path=config_path
            )

        resolved_dict = {}
        for key, value in config_dict.items():
            if isinstance(value, dict):
                resolved_dict[key] = self._resolve_paths(
                    value,
                    config_path=config_path,
                    parent_path=parent_path,
                )
            elif isinstance(value, str) and (
                "path_" in key.lower() or "location" in key.lower()
            ):
                path = Path(value)
                if not path.is_absolute():
                    path = (parent_path / path).resolve()

                resolved_dict[key] = str(path)
            else:
                resolved_dict[key] = value
        return resolved_dict

    def load_configuration(self, file_path: str) -> None:
        """
        Load and validate nested configuration.

        Parameters
        ----------
        file_path : str
            Path to YAML configuration file
        """

        config_path = Path(file_path).resolve()

        with open(file_path) as f:
            config_dict = yaml.safe_load(f)
        config_dict = self._resolve_paths(config_dict, config_path)

        if config_dict["config"] == "sensor":
            config_type = str(ConfigType.SENSOR.value)
            self._configs[config_type] = SensorConfig(**config_dict)
        elif config_dict["config"] == "process":
            config_type = str(ConfigType.PROCESS.value)
            self._configs[config_type] = ProcessConfig(**config_dict)

    def get_config(self, name: Literal["sensor", "process"]):
        """
        Return the specific config

        Parameters
        ----------
        name : str
            Either sensor or process

        Returns
        -------
        BaseConfig
            The requested config
        """
        return self._configs[name]

    def create_sensor_config(self):
        pass



import pandas as pd
import numpy as np
from typing import Literal, Union, Optional
from pathlib import Path
from neptoon.external.nmdb_data_collection import (
    NMDBDataAttacher,
)
from neptoon.calibration import (
    CalibrationConfiguration,
    CalibrationStation,
)
from neptoon.corrections.factory.build_corrections import (
    CorrectionBuilder,
    CorrectionFactory,
    CorrectionType,
    CorrectionTheory,
    CorrectNeutrons,
    NeutronUncertaintyCalculator,
)
from neptoon.config.configuration_input import SensorInfo
from neptoon.products.estimate_sm import NeutronsToSM
from neptoon.quality_control.data_validation_tables import (
    FormatCheck,
)
from neptoon.quality_control import (
    QualityAssessmentFlagBuilder,
    DataQualityAssessor,
)
from neptoon.visulisation.figures_handler import FigureHandler
from neptoon.io.save import SaveAndArchiveOutputs
from neptoon.data_prep.smoothing import SmoothData
from neptoon.data_prep.conversions import AbsoluteHumidityCreator
from neptoon.columns import ColumnInfo
from neptoon.logging import get_logger
from magazine import Magazine

core_logger = get_logger()


class CRNSDataHub:
    """
    The CRNSDataHub is used to manage the time series data throughout
    the processing steps. Some key features:

    - It stores a DataFrame for a site
    - As we progress through the steps, data can be added to the
      DataFrame and the shadow DataFrame's updated.

    Raw data is checked against the RawDataSchema which is a first line
    of defense against incorrectly formatted tables. Should a fail
    happen here data must be either reformatted using one of the
    provided routines or manually formatted to match the standard.
    """

    def __init__(
        self,
        crns_data_frame: pd.DataFrame,
        flags_data_frame: pd.DataFrame | None = None,
        sensor_info: SensorInfo | None = None,
        quality_assessor: DataQualityAssessor | None = None,
        validation: bool = True,
        calibration_samples_data: pd.DataFrame | None = None,
    ):
        """
        Inputs to the CRNSDataHub.

        Parameters
        ----------
        crns_data_frame : pd.DataFrame
            CRNS data in a dataframe format. It will be validated to
            ensure it has been formatted correctly.
        configuration_manager : ConfigurationManager, optional
            A ConfigurationManager instance storing configuration YAML
            information, by default None
        quality_assessor : SaQC
            SaQC object which is used for quality assessment. Used for
            the creation of flags to define poor data.
        validation : bool
            Toggle for whether to have continued validation of data
            tables during processing (see
            data_management>data_validation_tables.py for examples of
            tables being validated). These checks ensure data is
            correctly formatted for internal processing.
        calibration_samples_data : pd.DataFrame
            The sample data taken during the calibration campaign.
        """

        self._raw_data = crns_data_frame.copy()
        self._crns_data_frame = crns_data_frame
        self._flags_data_frame = flags_data_frame
        self._sensor_info = sensor_info
        self._validation = validation
        self._quality_assessor = quality_assessor
        self._calibration_samples_data = calibration_samples_data
        self._correction_factory = CorrectionFactory()
        self._correction_builder = CorrectionBuilder()
        self.calibrator = None
        self.figure_creator = None
        self.magazine_active = [Magazine.active if Magazine.active else False]

    @property
    def crns_data_frame(self):
        return self._crns_data_frame

    @crns_data_frame.setter
    def crns_data_frame(self, df: pd.DataFrame):
        self._crns_data_frame = df

    @property
    def flags_data_frame(self):
        return self._flags_data_frame

    @flags_data_frame.setter
    def flags_data_frame(self, df: pd.DataFrame):
        # TODO checks on df
        self._flags_data_frame = df

    @property
    def sensor_info(self):
        return self._sensor_info

    @sensor_info.setter
    def sensor_info(self, new_config: SensorInfo):
        self._sensor_info = new_config

    @property
    def validation(self):
        return self._validation

    @property
    def quality_assessor(self):
        return self._quality_assessor

    @quality_assessor.setter
    def quality_assessor(self, assessor: DataQualityAssessor):
        if isinstance(assessor, DataQualityAssessor):
            self._quality_assessor = assessor
        else:
            message = (
                f"{assessor} is not a DataQualityAssessor class. "
                "Cannot assign to self.quality_assessor"
            )
            core_logger.error(message)
            raise TypeError(message)

    @property
    def correction_factory(self):
        return self._correction_factory

    @property
    def calibration_samples_data(self):
        return self._calibration_samples_data

    @calibration_samples_data.setter
    def calibration_samples_data(self, data: pd.DataFrame):
        # TODO add verification
        self._calibration_samples_data = data

    @property
    def correction_builder(self):
        return self._correction_builder

    @correction_builder.setter
    def correction_builder(self, builder: CorrectionBuilder):
        self._correction_builder = builder

    def validate_dataframe(
        self,
        schema: str,
    ):
        """
        Validates the dataframe against a pandera schema See
        data_validation_table.py for schemas.

        Parameters
        ----------
        schema : str
            The name of the schema to use for the check.
        """

        if schema == "initial_check":
            tmpdf = self.crns_data_frame
            FormatCheck.validate(tmpdf, lazy=True)
        elif schema == "before_corrections_check":
            pass
        elif schema == "after_corrections_check":
            pass
        elif schema == "final_check":
            pass
        else:
            validation_error_message = (
                "Incorrect schema or table name given "
                "when validating the crns_data_frame"
            )
            core_logger.error(validation_error_message)
            print(validation_error_message)

    @Magazine.reporting(topic="NMDB")
    def attach_nmdb_data(
        self,
        station="JUNG",
        new_column_name=str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY),
        resolution="60",
        nmdb_table="revori",
        reference_value: int | None = None,
    ):
        """
        Utilises the NMDBDataAttacher class to attach NMDB incoming
        intensity data to the crns_data_frame. Collects data using
        www.NMDB.eu

        See NMDBDataAttacher documentation for more information.

        Parameters
        ----------
        station : str, optional
            The station to collect data from, by default "JUNG"
        new_column_name : str, optional
            The name of the column were data will be written to, by
            default "incoming_neutron_intensity"
        resolution : str, optional
            The resolution in minutes, by default "60"
        nmdb_table : str, optional
            The table to pull data from, by default "revori"
        reference_value : int, optional
            The reference value of the neutron monitor, if left as None
            it will use the value from the first data point in the time
            series.
        Report
        ------
        Neutron monitoring data was attached from NMDB.eu. The station
        used was {station} at a resolution of {resolution} minutes. The
        data table used was {nmdb_table}.
        """
        attacher = NMDBDataAttacher(
            data_frame=self.crns_data_frame, new_column_name=new_column_name
        )
        attacher.configure(
            station=station,
            reference_value=reference_value,
            resolution=resolution,
            nmdb_table=nmdb_table,
        )
        attacher.fetch_data()
        attacher.attach_data()
        self.crns_data_frame = attacher.return_data_frame()

    def add_quality_flags(
        self,
        custom_flags: QualityAssessmentFlagBuilder | None = None,
        add_check=None,
    ):
        """
        Add QualityChecks to undertake on the dataframe

        Parameters
        ----------
        custom_flags : QualityAssessmentFlagBuilder, optional
            user can build a QualityAssessmentFlagBuilder with checks
            and attach this as a whole, by default None
        add_check : Check, optional
            user can add individual Checks, or a list of Checks. These
            will be then added to the QualityAssessmentFlagBuilder, by
            default None
        """
        self.quality_assessor = DataQualityAssessor(
            data_frame=self.crns_data_frame
        )
        if custom_flags:
            self.quality_assessor.add_custom_flag_builder(custom_flags)

        if add_check:
            if isinstance(add_check, list):
                for check in add_check:
                    self.quality_assessor.add_quality_check(check)
            else:
                self.quality_assessor.add_quality_check(add_check)

    def apply_quality_flags(
        self,
    ):
        """
        Flags data based on quality assessment. A user can supply a
        QualityAssessmentFlagBuilder object that has been custom built,
        they can flag using the config file (if supplied), or they can
        choose a standard flagging routine.

        Everything is off by default so a user must choose.

        Parameters
        ----------
        custom_flags : QualityAssessmentFlagBuilder, optional
            A custom built set of Flags , by default None
        flags_from_config : bool, optional
            State if to conduct QA using config supplied configuration,
            by default False
        flags_default : str, optional
            A string representing a default version of flagging, by
            default None
        """
        self.quality_assessor.apply_quality_assessment()

        self.flags_data_frame = self.quality_assessor.return_flags_data_frame(
            current_flag_data_frame=self.flags_data_frame,
        )
        self.crns_data_frame = self.mask_flagged_data(
            data_frame=self.crns_data_frame
        )
        message = "Flagging of data complete using Custom Flags"
        core_logger.info(message)

    def select_correction(
        self,
        correction_type: CorrectionType | str = "empty",
        correction_theory: CorrectionTheory | None = None,
    ):
        """
        Method to select corrections to be applied to data.

        Individual corrections can be applied using a CorrectionType and
        CorrectionTheory. If a user assigns a CorrectionType without a
        CorrectionTheory, then the default correction for that
        CorrectionType is applied.

        Parameters
        ----------
        correction_type : CorrectionType, optional
            A CorrectionType, by default "empty"
        correction_theory : CorrectionTheory, optional
            A CorrectionTheory, by default None
        """

        correction = self.correction_factory.create_correction(
            correction_type=correction_type,
            correction_theory=correction_theory,
        )
        self.correction_builder.add_correction(correction=correction)

    def correct_neutrons(
        self,
    ):
        """
        Create correction factors as well as the corrected epithermal
        neutrons column.
        """
        corrector = CorrectNeutrons(
            crns_data_frame=self.crns_data_frame,
            correction_builder=self.correction_builder,
        )
        self.crns_data_frame = corrector.correct_neutrons()

    def create_neutron_uncertainty_bounds(self):
        """
        Create uncertainty bounds for statistical uncertainty of neutron
        count rates.
        """

        uncertainty = NeutronUncertaintyCalculator(
            data_frame=self.crns_data_frame
        )
        self.crns_data_frame = uncertainty.add_neutron_uncertainty_columns()

    @Magazine.reporting(topic="Data Preparation")
    def smooth_data(
        self,
        column_to_smooth: str,
        smooth_method: Literal[
            "rolling_mean", "savitsky_golay"
        ] = "rolling_mean",
        window: Optional[Union[int, str]] = 12,
        poly_order: int = 4,
        auto_update_final_col: bool = True,
    ):
        """
        Applies a smoothing method to a series of data in the
        crns_data_frame using the SmoothData class.

        A `column_to_smooth` attribute must be supplied, and should be
        written using the "str(ColumnInfo.Name.COLUMN)" format. The two
        most likely to be used are:

           - str(ColumnInfo.Name.SOIL_MOISTURE)
           - str(ColumnInfo.Name.EPI_NEUTRONS)

        If parameters are left as None, it uses defaults from SmoothData
        (i.e., rolling_mean, window size == 12).

        Parameters
        ----------
        column_to_smooth : str(ColumnInfo.Name.VALUE)
            The column in the crns_data_frame that needs to be smoothed.
            Automatically

        Report
        ------
        Data smoothing was done on {column_to_smooth}. This was done
        using {smooth_method} with a window of {window}.
        """
        series_to_smooth = pd.Series(self.crns_data_frame[column_to_smooth])
        smoother = SmoothData(
            data=series_to_smooth,
            column_to_smooth=column_to_smooth,
            smooth_method=smooth_method,
            window=window,
            poly_order=poly_order,
            auto_update_final_col=auto_update_final_col,
        )
        col_name = smoother.create_new_column_name()
        self.crns_data_frame[col_name] = smoother.apply_smoothing()

    @Magazine.reporting(topic="Calibration")
    def calibrate_station(
        self,
        config: CalibrationConfiguration = None,
    ):
        """
        Calibrate the sensor

        Parameters
        ----------
        config : CalibrationConfiguration, optional
            Config file which contains all the required info for
            calibration, by default None

        Raises
        ------
        ValueError
            When no calibration data provided

        Report
        ------
        Calibration was undertaken. The N0 number was calculated as
        {n0}, using the {config.neutron_conversion_method} method. From the samples, the average dry soil bulk density is
        {avg_dry_soil_bulk_density}, the average soil organic carbon is
        {avg_soil_organic_carbon}, and the average lattice water content
        is {avg_lattice_water}.
        """
        if self.calibration_samples_data is None:
            message = "No calibration_samples_data found. Cannot calibrate."
            core_logger.error(message)
            raise ValueError(message)
        if config is None:
            message = "No CalibrationConfiguration provided - using defaults"
            core_logger.info(message)
            config = CalibrationConfiguration()

        self.calibrator = CalibrationStation(
            calibration_data=self.calibration_samples_data,
            time_series_data=self.crns_data_frame,
            config=config,
        )
        n0 = int(self.calibrator.find_n0_value())
        avg_dry_soil_bulk_density = round(
            self.calibrator.calibrator.calib_data_object.list_of_profiles[
                0
            ].site_avg_bulk_density,
            4,
        )
        avg_soil_organic_carbon = round(
            self.calibrator.calibrator.calib_data_object.list_of_profiles[
                0
            ].site_avg_organic_carbon,
            4,
        )
        avg_lattice_water = round(
            self.calibrator.calibrator.calib_data_object.list_of_profiles[
                0
            ].site_avg_lattice_water,
        )
        self.sensor_info.N0 = n0
        self.sensor_info.avg_dry_soil_bulk_density = avg_dry_soil_bulk_density
        self.sensor_info.avg_lattice_water = avg_lattice_water
        self.sensor_info.avg_soil_organic_carbon = avg_soil_organic_carbon
        print(f"N0 number was calculated as {n0}")

    @Magazine.reporting(topic="Soil Moisture")
    def produce_soil_moisture_estimates(
        self,
        n0: float | None = None,
        conversion_theory: Literal[
            "desilets_etal_2010", "koehli_eta_2021"
        ] = "desilets_etal_2010",
        dry_soil_bulk_density: float | None = None,
        lattice_water: float | None = None,
        soil_organic_carbon: float | None = None,
        koehli_method_form: Literal[
            "Jan23_uranos",
            "Jan23_mcnpfull",
            "Mar12_atmprof",
            "Mar21_mcnp_drf",
            "Mar21_mcnp_ewin",
            "Mar21_uranos_drf",
            "Mar21_uranos_ewin",
            "Mar22_mcnp_drf_Jan",
            "Mar22_mcnp_ewin_gd",
            "Mar22_uranos_drf_gd",
            "Mar22_uranos_ewin_chi2",
            "Mar22_uranos_drf_h200m",
            "Aug08_mcnp_drf",
            "Aug08_mcnp_ewin",
            "Aug12_uranos_drf",
            "Aug12_uranos_ewin",
            "Aug13_uranos_atmprof",
            "Aug13_uranos_atmprof2",
        ] = "Mar21_uranos_drf",
    ):
        """
        Produces SM estimates with the NeutronsToSM class. If values for
        n0, dry_soil_bulk_density, lattice_water, or soil_organic_carbon
        are not supplied, the values are taken from the internal
        sensor_info class.

        Parameters
        ----------
        n0 : float, optional
            n0 calibration term, by default None
        dry_soil_bulk_density : float, optional
            given in g/cm3, by default None
        lattice_water : float, optional
            given as decimal percent e.g., 0.01, by default None
        soil_organic_carbon : float, optional
            Given as decimal percent, e.g., 0.001, by default None

        Report
        ------
        Soil moisture was estimated using an n0 of {default_params[n0]},
        a bulk density of {default_params[dry_soil_bulk_density]}, a
        lattice water content of {default_params[lattice_water]}, and a
        soil organic carbon content of
        {default_params[soil_organic_carbon]}
        """
        print("Producing soil moisture estimates.")
        # Create attributes for NeutronsToSM
        default_params = {
            "n0": self.sensor_info.N0,
            "dry_soil_bulk_density": self.sensor_info.avg_dry_soil_bulk_density,
            "lattice_water": self.sensor_info.avg_lattice_water,
            "soil_organic_carbon": self.sensor_info.avg_soil_organic_carbon,
        }
        provided_params = {
            "n0": n0,
            "dry_soil_bulk_density": dry_soil_bulk_density,
            "lattice_water": lattice_water,
            "soil_organic_carbon": soil_organic_carbon,
            "conversion_theory": conversion_theory,
            "koehli_method_form": koehli_method_form,
        }
        params = {k: v for k, v in provided_params.items() if v is not None}
        default_params.update(params)

        soil_moisture_calculator = NeutronsToSM(
            crns_data_frame=self.crns_data_frame, **default_params
        )
        soil_moisture_calculator.calculate_all_soil_moisture_data()
        self.crns_data_frame = soil_moisture_calculator.return_data_frame()

    def mask_flagged_data(self, data_frame: pd.DataFrame):
        """
        Returns a pd.DataFrame() where flagged data has been replaced
        with np.nan values
        """
        mask = self.flags_data_frame == "UNFLAGGED"
        data_frame[~mask] = np.nan
        return data_frame

    def prepare_static_values(self):
        """
        Attaches the static values from the SensorInfo Pydantic model as
        columns of values in the crns_data_frame.

        This method:
        1. Converts the Pydantic model to a dictionary
        2. Checks if each key already exists in the DataFrame
        3. Skips None values
        4. Adds the remaining values as new columns

        The method preserves existing column values if they are already
        present in the DataFrame to avoid accidental overwrites.
        """

        sensor_info_dict = self.sensor_info.model_dump()
        for key, value in sensor_info_dict.items():
            if key in self.crns_data_frame.columns:
                message = (
                    f"{key} already found in columns of crns_data_frame"
                    " when trying to add static values from sensor_info."
                    "This value from SensorInfo was not written to the"
                    " crns_data_frame."
                )
                core_logger.info(message)
                continue
            elif value is None:
                # TODO add skip here
                continue
            else:
                self.crns_data_frame[key] = value

    def prepare_additional_columns(self):
        """
        Prepares and adds additional columns required for processing.

        Such as:
           - absolute humidity
        """
        abs_hum_creator = AbsoluteHumidityCreator(
            data_frame=self.crns_data_frame
        )
        self.crns_data_frame = (
            abs_hum_creator.check_and_return_abs_hum_column()
        )

    def create_figures(
        self,
        create_all=True,
        ignore_sections=[],
        selected_figures=[],
        show_figures: bool = False,
    ):
        """
        Handles creating the figures using the FigureHandler.

        Parameters
        ----------
        create_all : bool, optional
            Default to create all figures in the
            FigureHandler._register, by default True
        ignore_sections : list, optional
            Ignore a whole topic section of figure names, by default []
        selected_figures : list, optional
            A list of the figures to be created if not using create_all.
            See FigureHandler._figure_registry for the names of possible
            figures, by default []
        show_figures : bool, optional
            Turn to False to not show Figures in the kernel, by default
            True
        """
        self.figure_creator = FigureHandler(
            data_frame=self.crns_data_frame,
            sensor_info=self.sensor_info,
            create_all=create_all,
            ignore_sections=ignore_sections,
            selected_figures=selected_figures,
            show_figures=show_figures,
        )
        self.figure_creator.create_figures()

    def save_data(
        self,
        folder_name: Union[str, None] = None,
        save_folder_location: Union[str, Path, None] = None,
        append_audit_log_hash_to_folder_name: bool = False,
        use_custom_column_names: bool = False,
        custom_column_names_dict: Union[dict, None] = None,
        append_timestamp: bool = True,
    ):
        """
        Saves the file to a specified location. It must contain the
        correct folder_path and file_name.

        Parameters
        ----------
        folder_path : str
            Path to the save folder
        file_name : str
            Name of the file
        """
        if folder_name is None:
            folder_name = self.sensor_info.name
        if save_folder_location is None:
            save_folder_location = Path.cwd()
        if self.calibrator:
            calib_df = self.calibrator.return_calibration_results_data_frame()
        else:
            calib_df = None

        self.saver = SaveAndArchiveOutputs(
            folder_name=folder_name,
            processed_data_frame=self.crns_data_frame,
            flag_data_frame=self.flags_data_frame,
            sensor_info=self.sensor_info,
            save_folder_location=save_folder_location,
            append_audit_log_hash_to_folder_name=append_audit_log_hash_to_folder_name,
            use_custom_column_names=use_custom_column_names,
            custom_column_names_dict=custom_column_names_dict,
            append_timestamp=append_timestamp,
            figure_handler=self.figure_creator,
            calib_df=calib_df,
            magazine_active=self.magazine_active,
        )
        self.saver.save_outputs()



"""
Functions related neutron corrections for changes to air humidity

Functions in this module:

    humidity_correction_rosolem2013
    saturation_vapour_pressure
    calc_absolute_humidity
    calc_vapour_pressure_from_dewpoint_temp
    calc_relative_humidity_from_temperature

"""

import numpy as np
from neptoon.logging import get_logger

core_logger = get_logger()


def humidity_correction_rosolem2013(
    absolute_humidity: float, reference_absolute_humidity: float = 0
):
    """
    Calculate the correction factor for neutron counts based on the
    difference in absolute humidity between current and reference
    conditions.

    Parameters
    ----------
    absolute_humidity : float
        Current absolute humidity in grams per cubic meter (g/m^3).
    reference_absolute_humidity : float
        Reference absolute humidity in grams per cubic meter (g/m^3).

    Returns
    -------
    c_humidity: float
        Correction factor to be multiplied with neutron counts.
    """
    c_humidity = 1 + 0.0054 * (absolute_humidity - reference_absolute_humidity)
    return c_humidity


def calc_absolute_humidity(vapour_pressure: float, temperature: float):
    """
    Calculate absolute humidity using air temperature and vapour
    pressure.

    Parameters
    ----------
    vapour_pressure : float
        Vapour pressure in pascals (hPa)
    temperature : float
        Air temperature in Celsius (C)

    Returns
    -------
    absolute_humidity: float
        Absolute humidity in grams per cubic meter (g/m^3)
    """
    absolute_humidity = (
        (vapour_pressure * 100) / (461.5 * (temperature + 273.15))
    ) * 1000
    return absolute_humidity


def calc_saturation_vapour_pressure(temperature: float):
    """
    Calculate saturation vapour pressure (hPA) using average temperature
    Can be used to calculate actual vapour pressure (hPA) if using dew
    point temperature

    Parameters
    ----------
    temperature : float
        temperature (C)

    Returns
    -------
    float
        saturation vapour pressure (hPA)
    """
    return 6.112 * np.exp((17.67 * temperature) / (243.5 + temperature))


def calc_vapour_pressure_from_dewpoint_temp(dewpoint_temp: float):
    """
    Calculate vapour pressure from the dewpoint temperature,
    particularly useful when using ERA5-Land data.

    Parameters
    ----------
    dewpoint_temp : float
        Dewpoint temperature in degrees Celsius (C)

    Returns
    -------
    vapour_pressure: float
        Vapour pressure in hectopascals (hPa).

    Example
    -------
    >>> calc_vapour_pressure_from_dewpoint_temp(12)
    1.40
    """
    vapour_pressure = (
        np.exp(
            (0.0707 * dewpoint_temp - 0.49299) / (1 + 0.00421 * dewpoint_temp)
        )
    ) / 10
    return vapour_pressure


def calc_relative_humidity_from_dewpoint_temperature(
    temperature: float, dewpoint_temperature: float
):
    """
    Calculate relative humidity from temperature and dewpoint temperature

    Parameters
    ----------
    temperature : float
        Temperature in Celsius (C)
    dewpoint_temperature : float
        Dewpoint temperature in Celsius (C)

    Returns
    -------
    relative_humidity: float
        Relative humidity (%)
    """
    relative_humidity = 100 * np.exp(
        (17.625 * 243.04 * (dewpoint_temperature - temperature))
        / ((243.04 + temperature) * (243.04 + dewpoint_temperature))
    )
    return relative_humidity


def calc_actual_vapour_pressure(
    saturation_vapour_pressure: float, relative_humidity: float
):
    """
    Calculates actual vapour pressure

    Parameters
    ----------
    saturation_vapour_pressure : float
        Saturation Vapour Pressure (hPa)
    relative_humidity : float
        Relative Humdity (%)

    Returns
    -------
    actual vapour pressure: float
        The actual vapour pressure (hPa)
    """
    return saturation_vapour_pressure * (relative_humidity / 100)



import pandas as pd
import numpy as np
from neptoon.logging import get_logger
from neptoon.columns import ColumnInfo
from .correction_classes import (
    Correction,
    CorrectionType,
    CorrectionTheory,
    IncomingIntensityCorrectionZreda2012,
    IncomingIntensityCorrectionHawdon2014,
    IncomingIntensityCorrectionMcJannetDesilets2023,
    HumidityCorrectionRosolem2013,
    PressureCorrectionZreda2012,
    AboveGroundBiomassCorrectionBaatz2015,
    AboveGroundBiomassCorrectionMorris2024,
)

core_logger = get_logger()


def calculate_poisson_uncertainty():
    pass


class CorrectionBuilder:
    """
    Staging place for the corrections as they are built. First a user
    adds a check using the add_check method.

    Parameters
    ----------

    corrections : dict
        dictionary which contains the corrections. The key is the
        correction_type assigned in each correction, the value is the
        correction itself.

    """

    def __init__(self):
        self.corrections = {}

    def add_correction(self, correction: Correction):
        """
        Adds a correction to the CorrectionBuilder

        Parameters
        ----------
        correction : Correction
            A Correction object
        """

        if isinstance(correction, Correction):
            correction_type = correction.correction_type
            self.corrections[correction_type] = correction
        else:
            message = f"{correction} is not a correction"
            core_logger.error(message)

    def remove_correction_by_type(self, correction_type: str):
        """
        Removes a correction from the CorrectionBuilder based on its
        type

        Parameters
        ----------
        correction_type : str
            The type of correction to be removed
        """
        if correction_type in self.corrections:
            self.corrections.pop(correction_type)
        else:
            raise ValueError(
                f"Correction type '{correction_type}' not found in the builder."
            )

    def get_corrections(self):
        """
        Returns the corrections stored in the builder
        """
        return self.corrections.values()


class CorrectNeutrons:
    """
    CorrectNeutrons class handles correcting neutrons for additional
    influences beyond soil moisture. It takes in a crns_data_frame which
    is a pd.DataFrame which stores the required data.

    Methods are available for staging a series of corrections which are
    applied to remove additional influences on the neutron signal. A
    user can add corrections individually, or create a CorrectionBuilder
    class seperately that has been pre-compiled and inject that into the
    CorrectNeutrons instance.

    Once the CorrectionBuilder has been appropriately staged with
    desired corrections, the correct_neutrons method will apply each
    correction, record the correction factor and create a corrected
    epithermal neutron count column with the correction factors applied.
    """

    def __init__(
        self,
        crns_data_frame: pd.DataFrame,
        correction_builder: CorrectionBuilder,
    ):
        """
        Attributes for using the CorrectNeutrons class

        Parameters
        ----------
        crns_data_frame : pd.DataFrame
            A DataFrame which contains the appropriate information to
            apply corrections.
        correction_builder : CorrectionBuilder
            Staging area for corrections. Can be built or supplied
            completed.
        """
        self._crns_data_frame = crns_data_frame
        self._correction_builder = correction_builder
        self._correction_columns = []

    @property
    def crns_data_frame(self):
        return self._crns_data_frame

    @crns_data_frame.setter
    def crns_data_frame(self, df: pd.DataFrame):
        # TODO add checks that it is df
        self._crns_data_frame = df

    @property
    def correction_builder(self):
        return self._correction_builder

    @correction_builder.setter
    def correction_builder(self, builder: CorrectionBuilder):
        if isinstance(builder, CorrectionBuilder):
            self._correction_builder = builder
        else:
            message = f"It appears {builder} is not a CorrectionBuilder object"
            core_logger.error(message)
            raise AttributeError

    @property
    def correction_columns(self):
        return self._correction_columns

    def add_correction(self, new_correction: Correction):
        """
        Add an invidual correction to the CorrectionBuilder

        Parameters
        ----------
        new_correction : Correction
            The new correction to apply
        """
        self.correction_builder.add_correction(new_correction)

    def add_correction_builder(
        self, new_correction_builder: CorrectionBuilder
    ):
        """
        Add a whole correction builder. Useful if a correction builder
        has been built somewhere in the code and you want to read it in.

        Parameters
        ----------
        new_correction_builder : CorrectionBuilder
            A pre-compiled correction builder.
        """
        self.correction_builder = new_correction_builder

    def create_correction_factors(self, df: pd.DataFrame):
        """
        Cycles through all the corrections in the CorrectionBuilder and
        applies them to the DataFrame. Returns the DataFrame with additional
        columns.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame which is prepared for correction.

        Returns
        -------
        pd.DataFrame
            DataFrame with additional columns applied during correction.
        """
        for correction in self.correction_builder.get_corrections():
            df = correction.apply(df)
            correction_column_name = (
                correction.get_correction_factor_column_name()
            )
            self.correction_columns.append(correction_column_name)

        return df

    def create_corrected_neutron_column(self, df):
        """
        Calculates the corrected neutron count rate after applying all
        the corrections.

        NOTE:

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame with the corrections applied and recorded in the
            columns

        Returns
        -------
        pd.DataFrame
            DataFrame with the corrected epithermal neutron count
            recorded in a column
        """
        df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)] = df[
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL)
        ]
        for column_name in self.correction_columns:
            df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)] = (
                df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)]
                * df[column_name]
            )
        return df

    def correct_neutrons(self):
        """
        Corrects neutrons using the CorrectionBuilder. Returns the
        DataFrame.

        Returns
        -------
        df: pd.DataFrame
            DataFrame returned with additional columns.
        """
        df = self.create_correction_factors(self.crns_data_frame)
        df = self.create_corrected_neutron_column(df)

        return df


class NeutronUncertaintyCalculator:
    """Calculates statistical uncertainty of neutron counts"""

    def __init__(self, data_frame: pd.DataFrame):
        """
        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame with neutron counts
        """
        self.data_frame = data_frame

    def add_neutron_uncertainty_columns(
        self,
    ):
        """
        Creates uncertainty neutron columns

        Returns
        -------
        pd.DataFrame
            Output DataFrame with uncertainty columns
        """
        self.data_frame = self.calculate_neutron_count_uncertainty(
            self.data_frame
        )
        self.data_frame = self.calculate_neutron_count_bounds(self.data_frame)
        return self.data_frame

    def calculate_neutron_count_uncertainty(
        self,
        df,
    ):
        """
        Adds a column to the DataFrame with the statistical error rate
        of neutrons.

        This method computes the statistical error of neutron counts
        based on Poisson statistics, where the error is the square root
        of the count. It then calculates this error as a proportion of
        the original count and applies it to the corrected neutron
        count.

        The result is stored in a new column in the DataFrame.
        """

        df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY)] = (
            1
            / np.sqrt(df[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW)])
            * df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)]
        )
        return df

    def calculate_neutron_count_bounds(
        self,
        df,
    ):
        """
        Calculates and adds upper and lower bounds for corrected neutron
        counts to the DataFrame.

        This method computes the upper and lower bounds of the corrected
        epithermal neutron counts based on the previously calculated
        uncertainty. These bounds represent a confidence interval around
        the corrected neutron count.

        Two new columns are added to the DataFrame:
            1. Upper bound of the corrected neutron count
            2. Lower bound of the corrected neutron count
        """

        # Calculate upper bound of neutron count
        df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UPPER_COUNT)] = (
            df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)]
            + df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY)]
        )

        # Calculate lower bound of neutron count
        df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_LOWER_COUNT)] = (
            df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)]
            - df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY)]
        )
        return df


class CorrectionFactory:
    """
    Used inside the CRNSDataHub when selecting which corrections to
    apply. Creating a correction involves providing a CorrectionType and
    (optionally) a CorrectionTheory. The factory will then use
    information from the SiteInformation object to create a Correction
    with the appropriate information provided. Additionally a user can
    register a custom correction for use in processing.
    """

    def __init__(self):
        self.custom_corrections = {}

    def create_correction(
        self,
        correction_type: CorrectionType,
        correction_theory: CorrectionTheory = None,
    ):
        """
        Creates a particular Correction. CorrectionType and
        CorrectionTheory enums are used for selection. If
        CorrectionTheory is left empty the default correction is
        selected.

        Parameters
        ----------
        correction_type : CorrectionType
            The correction type to stage can be:
                - CorrectionType.INCOMING_INTENSITY
                - CorrectionType.ABOVE_GROUND_BIOMASS
                - CorrectionType.PRESSURE
                - CorrectionType.HUMIDITY
                - CorrectionType.CUSTOM
        correction_theory : CorrectionTheory, optional
            The theory to apply leave blank to use the default format,
            otherwise see CorrectionTheory for options, by default None

        Returns
        -------
        Correction
            Returns a correction object with the site specific values
            read in from the SiteInfo class
        """
        if (correction_type, correction_theory) in self.custom_corrections:
            return self.custom_corrections[
                (correction_type, correction_theory)
            ](
                correction_type=correction_type,
            )
        if correction_type == CorrectionType.ABOVE_GROUND_BIOMASS:
            return self.create_biomass_correction(
                correction_theory=correction_theory
            )
        if correction_type == CorrectionType.INCOMING_INTENSITY:
            return self.create_intensity_correction(
                correction_theory=correction_theory
            )
        if correction_type == CorrectionType.PRESSURE:
            return self.create_pressure_correction(
                correction_theory=correction_theory
            )
        if correction_type == CorrectionType.HUMIDITY:
            return self.create_humidity_correction(
                correction_theory=correction_theory
            )

    def create_intensity_correction(self, correction_theory: CorrectionTheory):
        """
        Internal method for selecting the incoming neutron intensity
        correction to use. If no CorrectionTheory supplied it will use
        the default.

        Parameters
        ----------
        correction_theory : CorrectionTheory
            The CorrectionTheory to use

        Returns
        -------
        Correction
            Intensity correction with values filled in.
        """
        if correction_theory is None:
            return IncomingIntensityCorrectionHawdon2014()
        elif correction_theory == CorrectionTheory.ZREDA_2012:
            return IncomingIntensityCorrectionZreda2012()
        elif correction_theory == CorrectionTheory.HAWDON_2014:
            return IncomingIntensityCorrectionHawdon2014()
        elif correction_theory == CorrectionTheory.MCJANNET_DESILETS_2023:
            return IncomingIntensityCorrectionMcJannetDesilets2023()

    def create_biomass_correction(self, correction_theory: CorrectionTheory):
        """
        Internal method for selecting the above ground biomass
        correction to use. If no CorrectionTheory supplied it will use
        the default.

        NOTE:

        Parameters
        ----------
        correction_theory : CorrectionTheory
            The CorrectionTheory to use

        Returns
        -------
        Correction
            Above Ground Biomass correction with values filled in.
        """
        if correction_theory is None:
            return AboveGroundBiomassCorrectionMorris2024
        elif correction_theory == CorrectionTheory.BAATZ_2015:
            return AboveGroundBiomassCorrectionBaatz2015
        elif correction_theory == CorrectionTheory.MORRIS_2024:
            return AboveGroundBiomassCorrectionMorris2024

    def create_pressure_correction(self, correction_theory: CorrectionTheory):
        """
        Internal method for selecting the correct pressure correction to
        use.

        Parameters
        ----------
        correction_theory : CorrectionTheory
            The CorrectionTheory to apply

        Returns
        -------
        Correction
            Returns a pressure Correction with values inserted.
        """
        correction_theory = (
            correction_theory  # TODO remove this or leave it for consistency?
        )
        return PressureCorrectionZreda2012()

    def create_humidity_correction(
        self, correction_theory: CorrectionTheory = None
    ):
        """
        Internal method for selecting the correct humidity correction to
        use

        Parameters
        ----------
        correction_theory : CorrectionTheory
            The CorrectionTheory to apply

        Returns
        -------
        Correction
            Returns the Correction
        """
        if correction_theory is None:
            return HumidityCorrectionRosolem2013()
        elif correction_theory == CorrectionTheory.ROSOLEM_2013:
            return HumidityCorrectionRosolem2013()

    def register_custom_correction(
        self, correction_type: CorrectionType, theory: str, correction_class
    ):
        """
        Used to register a custom correction theory. Without
        registration the correction cannot be used.

        The correction_class must be an object which inherits the
        Correction class and follows the same principles as other
        Correction type objects. This means it requires an apply method,
        which takes in a dataframe and returns a dataframe along with
        updates calulcated along the way.

        The key difference of a custom Correction class is that it must
        take a SiteInformation object. The logic under apply() will then
        use this SiteInformation object to collect relevant information
        for processing.

        Parameters
        ----------
        correction_type : CorrectionType
            The type of correction being registered
        theory : str
            A name for the custom correction
        correction_class : Correction
            A custom Correction class. See documentation on how to make
            this.

        Raises
        ------
        ValueError
            If the correction class is not a Correction type object it
            throws an error
        """
        if not issubclass(correction_class, Correction):
            message = "Custom correction must inherit from Correction class"
            core_logger.error(message)
            raise ValueError(message)
        self.custom_corrections[(correction_type, theory)] = correction_class



from typing import Dict, List, Set

# A set of classifier names
sorted_classifiers: List[str] = [
    "Development Status :: 1 - Planning",
    "Development Status :: 2 - Pre-Alpha",
    "Development Status :: 3 - Alpha",
    "Development Status :: 4 - Beta",
    "Development Status :: 5 - Production/Stable",
    "Development Status :: 6 - Mature",
    "Development Status :: 7 - Inactive",
    "Environment :: Console",
    "Environment :: Console :: Curses",
    "Environment :: Console :: Framebuffer",
    "Environment :: Console :: Newt",
    "Environment :: Console :: svgalib",
    "Environment :: GPU",
    "Environment :: GPU :: NVIDIA CUDA",
    "Environment :: GPU :: NVIDIA CUDA :: 1.0",
    "Environment :: GPU :: NVIDIA CUDA :: 1.1",
    "Environment :: GPU :: NVIDIA CUDA :: 2.0",
    "Environment :: GPU :: NVIDIA CUDA :: 2.1",
    "Environment :: GPU :: NVIDIA CUDA :: 2.2",
    "Environment :: GPU :: NVIDIA CUDA :: 2.3",
    "Environment :: GPU :: NVIDIA CUDA :: 3.0",
    "Environment :: GPU :: NVIDIA CUDA :: 3.1",
    "Environment :: GPU :: NVIDIA CUDA :: 3.2",
    "Environment :: GPU :: NVIDIA CUDA :: 4.0",
    "Environment :: GPU :: NVIDIA CUDA :: 4.1",
    "Environment :: GPU :: NVIDIA CUDA :: 4.2",
    "Environment :: GPU :: NVIDIA CUDA :: 5.0",
    "Environment :: GPU :: NVIDIA CUDA :: 5.5",
    "Environment :: GPU :: NVIDIA CUDA :: 6.0",
    "Environment :: GPU :: NVIDIA CUDA :: 6.5",
    "Environment :: GPU :: NVIDIA CUDA :: 7.0",
    "Environment :: GPU :: NVIDIA CUDA :: 7.5",
    "Environment :: GPU :: NVIDIA CUDA :: 8.0",
    "Environment :: GPU :: NVIDIA CUDA :: 9.0",
    "Environment :: GPU :: NVIDIA CUDA :: 9.1",
    "Environment :: GPU :: NVIDIA CUDA :: 9.2",
    "Environment :: GPU :: NVIDIA CUDA :: 10.0",
    "Environment :: GPU :: NVIDIA CUDA :: 10.1",
    "Environment :: GPU :: NVIDIA CUDA :: 10.2",
    "Environment :: GPU :: NVIDIA CUDA :: 11",
    "Environment :: GPU :: NVIDIA CUDA :: 11.0",
    "Environment :: GPU :: NVIDIA CUDA :: 11.1",
    "Environment :: GPU :: NVIDIA CUDA :: 11.2",
    "Environment :: GPU :: NVIDIA CUDA :: 11.3",
    "Environment :: GPU :: NVIDIA CUDA :: 11.4",
    "Environment :: GPU :: NVIDIA CUDA :: 11.5",
    "Environment :: GPU :: NVIDIA CUDA :: 11.6",
    "Environment :: GPU :: NVIDIA CUDA :: 11.7",
    "Environment :: GPU :: NVIDIA CUDA :: 11.8",
    "Environment :: GPU :: NVIDIA CUDA :: 12",
    "Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.0",
    "Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.1",
    "Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.2",
    "Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.3",
    "Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.4",
    "Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.5",
    "Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.6",
    "Environment :: Handhelds/PDA's",
    "Environment :: MacOS X",
    "Environment :: MacOS X :: Aqua",
    "Environment :: MacOS X :: Carbon",
    "Environment :: MacOS X :: Cocoa",
    "Environment :: No Input/Output (Daemon)",
    "Environment :: OpenStack",
    "Environment :: Other Environment",
    "Environment :: Plugins",
    "Environment :: Web Environment",
    "Environment :: Web Environment :: Buffet",
    "Environment :: Web Environment :: Mozilla",
    "Environment :: Web Environment :: ToscaWidgets",
    "Environment :: WebAssembly",
    "Environment :: WebAssembly :: Emscripten",
    "Environment :: WebAssembly :: WASI",
    "Environment :: Win32 (MS Windows)",
    "Environment :: X11 Applications",
    "Environment :: X11 Applications :: GTK",
    "Environment :: X11 Applications :: Gnome",
    "Environment :: X11 Applications :: KDE",
    "Environment :: X11 Applications :: Qt",
    "Framework :: AWS CDK",
    "Framework :: AWS CDK :: 1",
    "Framework :: AWS CDK :: 2",
    "Framework :: AiiDA",
    "Framework :: Ansible",
    "Framework :: AnyIO",
    "Framework :: Apache Airflow",
    "Framework :: Apache Airflow :: Provider",
    "Framework :: AsyncIO",
    "Framework :: BEAT",
    "Framework :: BFG",
    "Framework :: Bob",
    "Framework :: Bottle",
    "Framework :: Buildout",
    "Framework :: Buildout :: Extension",
    "Framework :: Buildout :: Recipe",
    "Framework :: CastleCMS",
    "Framework :: CastleCMS :: Theme",
    "Framework :: Celery",
    "Framework :: Chandler",
    "Framework :: CherryPy",
    "Framework :: CubicWeb",
    "Framework :: Dash",
    "Framework :: Datasette",
    "Framework :: Django",
    "Framework :: Django :: 1",
    "Framework :: Django :: 1.4",
    "Framework :: Django :: 1.5",
    "Framework :: Django :: 1.6",
    "Framework :: Django :: 1.7",
    "Framework :: Django :: 1.8",
    "Framework :: Django :: 1.9",
    "Framework :: Django :: 1.10",
    "Framework :: Django :: 1.11",
    "Framework :: Django :: 2",
    "Framework :: Django :: 2.0",
    "Framework :: Django :: 2.1",
    "Framework :: Django :: 2.2",
    "Framework :: Django :: 3",
    "Framework :: Django :: 3.0",
    "Framework :: Django :: 3.1",
    "Framework :: Django :: 3.2",
    "Framework :: Django :: 4",
    "Framework :: Django :: 4.0",
    "Framework :: Django :: 4.1",
    "Framework :: Django :: 4.2",
    "Framework :: Django :: 5",
    "Framework :: Django :: 5.0",
    "Framework :: Django :: 5.1",
    "Framework :: Django :: 5.2",
    "Framework :: Django CMS",
    "Framework :: Django CMS :: 3.4",
    "Framework :: Django CMS :: 3.5",
    "Framework :: Django CMS :: 3.6",
    "Framework :: Django CMS :: 3.7",
    "Framework :: Django CMS :: 3.8",
    "Framework :: Django CMS :: 3.9",
    "Framework :: Django CMS :: 3.10",
    "Framework :: Django CMS :: 3.11",
    "Framework :: Django CMS :: 4.0",
    "Framework :: Django CMS :: 4.1",
    "Framework :: Django CMS :: 5.0",
    "Framework :: FastAPI",
    "Framework :: Flake8",
    "Framework :: Flask",
    "Framework :: Hatch",
    "Framework :: Hypothesis",
    "Framework :: IDLE",
    "Framework :: IPython",
    "Framework :: Jupyter",
    "Framework :: Jupyter :: JupyterLab",
    "Framework :: Jupyter :: JupyterLab :: 1",
    "Framework :: Jupyter :: JupyterLab :: 2",
    "Framework :: Jupyter :: JupyterLab :: 3",
    "Framework :: Jupyter :: JupyterLab :: 4",
    "Framework :: Jupyter :: JupyterLab :: Extensions",
    "Framework :: Jupyter :: JupyterLab :: Extensions :: Mime Renderers",
    "Framework :: Jupyter :: JupyterLab :: Extensions :: Prebuilt",
    "Framework :: Jupyter :: JupyterLab :: Extensions :: Themes",
    "Framework :: Kedro",
    "Framework :: Lektor",
    "Framework :: Masonite",
    "Framework :: Matplotlib",
    "Framework :: MkDocs",
    "Framework :: Nengo",
    "Framework :: Odoo",
    "Framework :: Odoo :: 8.0",
    "Framework :: Odoo :: 9.0",
    "Framework :: Odoo :: 10.0",
    "Framework :: Odoo :: 11.0",
    "Framework :: Odoo :: 12.0",
    "Framework :: Odoo :: 13.0",
    "Framework :: Odoo :: 14.0",
    "Framework :: Odoo :: 15.0",
    "Framework :: Odoo :: 16.0",
    "Framework :: Odoo :: 17.0",
    "Framework :: Odoo :: 18.0",
    "Framework :: OpenTelemetry",
    "Framework :: OpenTelemetry :: Distros",
    "Framework :: OpenTelemetry :: Exporters",
    "Framework :: OpenTelemetry :: Instrumentations",
    "Framework :: Opps",
    "Framework :: Paste",
    "Framework :: Pelican",
    "Framework :: Pelican :: Plugins",
    "Framework :: Pelican :: Themes",
    "Framework :: Plone",
    "Framework :: Plone :: 3.2",
    "Framework :: Plone :: 3.3",
    "Framework :: Plone :: 4.0",
    "Framework :: Plone :: 4.1",
    "Framework :: Plone :: 4.2",
    "Framework :: Plone :: 4.3",
    "Framework :: Plone :: 5.0",
    "Framework :: Plone :: 5.1",
    "Framework :: Plone :: 5.2",
    "Framework :: Plone :: 5.3",
    "Framework :: Plone :: 6.0",
    "Framework :: Plone :: 6.1",
    "Framework :: Plone :: 6.2",
    "Framework :: Plone :: Addon",
    "Framework :: Plone :: Core",
    "Framework :: Plone :: Distribution",
    "Framework :: Plone :: Theme",
    "Framework :: PySimpleGUI",
    "Framework :: PySimpleGUI :: 4",
    "Framework :: PySimpleGUI :: 5",
    "Framework :: Pycsou",
    "Framework :: Pydantic",
    "Framework :: Pydantic :: 1",
    "Framework :: Pydantic :: 2",
    "Framework :: Pylons",
    "Framework :: Pyodide",
    "Framework :: Pyramid",
    "Framework :: Pytest",
    "Framework :: Review Board",
    "Framework :: Robot Framework",
    "Framework :: Robot Framework :: Library",
    "Framework :: Robot Framework :: Tool",
    "Framework :: Scrapy",
    "Framework :: Setuptools Plugin",
    "Framework :: Sphinx",
    "Framework :: Sphinx :: Domain",
    "Framework :: Sphinx :: Extension",
    "Framework :: Sphinx :: Theme",
    "Framework :: Trac",
    "Framework :: Trio",
    "Framework :: Tryton",
    "Framework :: TurboGears",
    "Framework :: TurboGears :: Applications",
    "Framework :: TurboGears :: Widgets",
    "Framework :: Twisted",
    "Framework :: Wagtail",
    "Framework :: Wagtail :: 1",
    "Framework :: Wagtail :: 2",
    "Framework :: Wagtail :: 3",
    "Framework :: Wagtail :: 4",
    "Framework :: Wagtail :: 5",
    "Framework :: Wagtail :: 6",
    "Framework :: Wagtail :: 7",
    "Framework :: ZODB",
    "Framework :: Zope",
    "Framework :: Zope2",
    "Framework :: Zope3",
    "Framework :: Zope :: 2",
    "Framework :: Zope :: 3",
    "Framework :: Zope :: 4",
    "Framework :: Zope :: 5",
    "Framework :: aiohttp",
    "Framework :: cocotb",
    "Framework :: napari",
    "Framework :: tox",
    "Intended Audience :: Customer Service",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: End Users/Desktop",
    "Intended Audience :: Financial and Insurance Industry",
    "Intended Audience :: Healthcare Industry",
    "Intended Audience :: Information Technology",
    "Intended Audience :: Legal Industry",
    "Intended Audience :: Manufacturing",
    "Intended Audience :: Other Audience",
    "Intended Audience :: Religion",
    "Intended Audience :: Science/Research",
    "Intended Audience :: System Administrators",
    "Intended Audience :: Telecommunications Industry",
    "License :: Aladdin Free Public License (AFPL)",
    "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
    "License :: CeCILL-B Free Software License Agreement (CECILL-B)",
    "License :: CeCILL-C Free Software License Agreement (CECILL-C)",
    "License :: DFSG approved",
    "License :: Eiffel Forum License (EFL)",
    "License :: Free For Educational Use",
    "License :: Free For Home Use",
    "License :: Free To Use But Restricted",
    "License :: Free for non-commercial use",
    "License :: Freely Distributable",
    "License :: Freeware",
    "License :: GUST Font License 1.0",
    "License :: GUST Font License 2006-09-30",
    "License :: Netscape Public License (NPL)",
    "License :: Nokia Open Source License (NOKOS)",
    "License :: OSI Approved",
    "License :: OSI Approved :: Academic Free License (AFL)",
    "License :: OSI Approved :: Apache Software License",
    "License :: OSI Approved :: Apple Public Source License",
    "License :: OSI Approved :: Artistic License",
    "License :: OSI Approved :: Attribution Assurance License",
    "License :: OSI Approved :: BSD License",
    "License :: OSI Approved :: Blue Oak Model License (BlueOak-1.0.0)",
    "License :: OSI Approved :: Boost Software License 1.0 (BSL-1.0)",
    "License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)",
    "License :: OSI Approved :: CMU License (MIT-CMU)",
    "License :: OSI Approved :: Common Development and Distribution License 1.0 (CDDL-1.0)",
    "License :: OSI Approved :: Common Public License",
    "License :: OSI Approved :: Eclipse Public License 1.0 (EPL-1.0)",
    "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
    "License :: OSI Approved :: Educational Community License, Version 2.0 (ECL-2.0)",
    "License :: OSI Approved :: Eiffel Forum License",
    "License :: OSI Approved :: European Union Public Licence 1.0 (EUPL 1.0)",
    "License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)",
    "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
    "License :: OSI Approved :: GNU Affero General Public License v3",
    "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
    "License :: OSI Approved :: GNU Free Documentation License (FDL)",
    "License :: OSI Approved :: GNU General Public License (GPL)",
    "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
    "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
    "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
    "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
    "License :: OSI Approved :: Historical Permission Notice and Disclaimer (HPND)",
    "License :: OSI Approved :: IBM Public License",
    "License :: OSI Approved :: ISC License (ISCL)",
    "License :: OSI Approved :: MIT License",
    "License :: OSI Approved :: MIT No Attribution License (MIT-0)",
    "License :: OSI Approved :: MirOS License (MirOS)",
    "License :: OSI Approved :: Motosoto License",
    "License :: OSI Approved :: Mozilla Public License 1.0 (MPL)",
    "License :: OSI Approved :: Mozilla Public License 1.1 (MPL 1.1)",
    "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
    "License :: OSI Approved :: Mulan Permissive Software License v2 (MulanPSL-2.0)",
    "License :: OSI Approved :: NASA Open Source Agreement v1.3 (NASA-1.3)",
    "License :: OSI Approved :: Nethack General Public License",
    "License :: OSI Approved :: Nokia Open Source License",
    "License :: OSI Approved :: Open Group Test Suite License",
    "License :: OSI Approved :: Open Software License 3.0 (OSL-3.0)",
    "License :: OSI Approved :: PostgreSQL License",
    "License :: OSI Approved :: Python License (CNRI Python License)",
    "License :: OSI Approved :: Python Software Foundation License",
    "License :: OSI Approved :: Qt Public License (QPL)",
    "License :: OSI Approved :: Ricoh Source Code Public License",
    "License :: OSI Approved :: SIL Open Font License 1.1 (OFL-1.1)",
    "License :: OSI Approved :: Sleepycat License",
    "License :: OSI Approved :: Sun Public License",
    "License :: OSI Approved :: The Unlicense (Unlicense)",
    "License :: OSI Approved :: Universal Permissive License (UPL)",
    "License :: OSI Approved :: University of Illinois/NCSA Open Source License",
    "License :: OSI Approved :: Vovida Software License 1.0",
    "License :: OSI Approved :: W3C License",
    "License :: OSI Approved :: Zero-Clause BSD (0BSD)",
    "License :: OSI Approved :: Zope Public License",
    "License :: OSI Approved :: zlib/libpng License",
    "License :: Other/Proprietary License",
    "License :: Public Domain",
    "License :: Repoze Public License",
    "Natural Language :: Afrikaans",
    "Natural Language :: Arabic",
    "Natural Language :: Armenian",
    "Natural Language :: Basque",
    "Natural Language :: Bengali",
    "Natural Language :: Bosnian",
    "Natural Language :: Bulgarian",
    "Natural Language :: Cantonese",
    "Natural Language :: Catalan",
    "Natural Language :: Catalan (Valencian)",
    "Natural Language :: Chinese (Simplified)",
    "Natural Language :: Chinese (Traditional)",
    "Natural Language :: Croatian",
    "Natural Language :: Czech",
    "Natural Language :: Danish",
    "Natural Language :: Dutch",
    "Natural Language :: English",
    "Natural Language :: Esperanto",
    "Natural Language :: Estonian",
    "Natural Language :: Finnish",
    "Natural Language :: French",
    "Natural Language :: Galician",
    "Natural Language :: Georgian",
    "Natural Language :: German",
    "Natural Language :: Greek",
    "Natural Language :: Hebrew",
    "Natural Language :: Hindi",
    "Natural Language :: Hungarian",
    "Natural Language :: Icelandic",
    "Natural Language :: Indonesian",
    "Natural Language :: Irish",
    "Natural Language :: Italian",
    "Natural Language :: Japanese",
    "Natural Language :: Javanese",
    "Natural Language :: Korean",
    "Natural Language :: Latin",
    "Natural Language :: Latvian",
    "Natural Language :: Lithuanian",
    "Natural Language :: Macedonian",
    "Natural Language :: Malay",
    "Natural Language :: Marathi",
    "Natural Language :: Nepali",
    "Natural Language :: Norwegian",
    "Natural Language :: Panjabi",
    "Natural Language :: Persian",
    "Natural Language :: Polish",
    "Natural Language :: Portuguese",
    "Natural Language :: Portuguese (Brazilian)",
    "Natural Language :: Romanian",
    "Natural Language :: Russian",
    "Natural Language :: Serbian",
    "Natural Language :: Slovak",
    "Natural Language :: Slovenian",
    "Natural Language :: Spanish",
    "Natural Language :: Swedish",
    "Natural Language :: Tamil",
    "Natural Language :: Telugu",
    "Natural Language :: Thai",
    "Natural Language :: Tibetan",
    "Natural Language :: Turkish",
    "Natural Language :: Ukrainian",
    "Natural Language :: Urdu",
    "Natural Language :: Vietnamese",
    "Natural Language :: Yiddish",
    "Operating System :: Android",
    "Operating System :: BeOS",
    "Operating System :: MacOS",
    "Operating System :: MacOS :: MacOS 9",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: Microsoft",
    "Operating System :: Microsoft :: MS-DOS",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: Microsoft :: Windows :: Windows 3.1 or Earlier",
    "Operating System :: Microsoft :: Windows :: Windows 7",
    "Operating System :: Microsoft :: Windows :: Windows 8",
    "Operating System :: Microsoft :: Windows :: Windows 8.1",
    "Operating System :: Microsoft :: Windows :: Windows 10",
    "Operating System :: Microsoft :: Windows :: Windows 11",
    "Operating System :: Microsoft :: Windows :: Windows 95/98/2000",
    "Operating System :: Microsoft :: Windows :: Windows CE",
    "Operating System :: Microsoft :: Windows :: Windows NT/2000",
    "Operating System :: Microsoft :: Windows :: Windows Server 2003",
    "Operating System :: Microsoft :: Windows :: Windows Server 2008",
    "Operating System :: Microsoft :: Windows :: Windows Vista",
    "Operating System :: Microsoft :: Windows :: Windows XP",
    "Operating System :: OS Independent",
    "Operating System :: OS/2",
    "Operating System :: Other OS",
    "Operating System :: PDA Systems",
    "Operating System :: POSIX",
    "Operating System :: POSIX :: AIX",
    "Operating System :: POSIX :: BSD",
    "Operating System :: POSIX :: BSD :: BSD/OS",
    "Operating System :: POSIX :: BSD :: FreeBSD",
    "Operating System :: POSIX :: BSD :: NetBSD",
    "Operating System :: POSIX :: BSD :: OpenBSD",
    "Operating System :: POSIX :: GNU Hurd",
    "Operating System :: POSIX :: HP-UX",
    "Operating System :: POSIX :: IRIX",
    "Operating System :: POSIX :: Linux",
    "Operating System :: POSIX :: Other",
    "Operating System :: POSIX :: SCO",
    "Operating System :: POSIX :: SunOS/Solaris",
    "Operating System :: PalmOS",
    "Operating System :: RISC OS",
    "Operating System :: Unix",
    "Operating System :: iOS",
    "Programming Language :: APL",
    "Programming Language :: ASP",
    "Programming Language :: Ada",
    "Programming Language :: Assembly",
    "Programming Language :: Awk",
    "Programming Language :: Basic",
    "Programming Language :: C",
    "Programming Language :: C#",
    "Programming Language :: C++",
    "Programming Language :: Cold Fusion",
    "Programming Language :: Cython",
    "Programming Language :: D",
    "Programming Language :: Delphi/Kylix",
    "Programming Language :: Dylan",
    "Programming Language :: Eiffel",
    "Programming Language :: Emacs-Lisp",
    "Programming Language :: Erlang",
    "Programming Language :: Euler",
    "Programming Language :: Euphoria",
    "Programming Language :: F#",
    "Programming Language :: Forth",
    "Programming Language :: Fortran",
    "Programming Language :: Go",
    "Programming Language :: Haskell",
    "Programming Language :: Hy",
    "Programming Language :: Java",
    "Programming Language :: JavaScript",
    "Programming Language :: Kotlin",
    "Programming Language :: Lisp",
    "Programming Language :: Logo",
    "Programming Language :: Lua",
    "Programming Language :: ML",
    "Programming Language :: Modula",
    "Programming Language :: OCaml",
    "Programming Language :: Object Pascal",
    "Programming Language :: Objective C",
    "Programming Language :: Other",
    "Programming Language :: Other Scripting Engines",
    "Programming Language :: PHP",
    "Programming Language :: PL/SQL",
    "Programming Language :: PROGRESS",
    "Programming Language :: Pascal",
    "Programming Language :: Perl",
    "Programming Language :: Pike",
    "Programming Language :: Pliant",
    "Programming Language :: Prolog",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2 :: Only",
    "Programming Language :: Python :: 2.3",
    "Programming Language :: Python :: 2.4",
    "Programming Language :: Python :: 2.5",
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.0",
    "Programming Language :: Python :: 3.1",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "Programming Language :: Python :: 3.14",
    "Programming Language :: Python :: 3.15",
    "Programming Language :: Python :: Free Threading",
    "Programming Language :: Python :: Free Threading :: 1 - Unstable",
    "Programming Language :: Python :: Free Threading :: 2 - Beta",
    "Programming Language :: Python :: Free Threading :: 3 - Stable",
    "Programming Language :: Python :: Free Threading :: 4 - Resilient",
    "Programming Language :: Python :: Implementation",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: GraalPy",
    "Programming Language :: Python :: Implementation :: IronPython",
    "Programming Language :: Python :: Implementation :: Jython",
    "Programming Language :: Python :: Implementation :: MicroPython",
    "Programming Language :: Python :: Implementation :: PyPy",
    "Programming Language :: Python :: Implementation :: Stackless",
    "Programming Language :: R",
    "Programming Language :: REBOL",
    "Programming Language :: Rexx",
    "Programming Language :: Ruby",
    "Programming Language :: Rust",
    "Programming Language :: SQL",
    "Programming Language :: Scheme",
    "Programming Language :: Simula",
    "Programming Language :: Smalltalk",
    "Programming Language :: Tcl",
    "Programming Language :: Unix Shell",
    "Programming Language :: Visual Basic",
    "Programming Language :: XBasic",
    "Programming Language :: YACC",
    "Programming Language :: Zope",
    "Topic :: Adaptive Technologies",
    "Topic :: Artistic Software",
    "Topic :: Communications",
    "Topic :: Communications :: BBS",
    "Topic :: Communications :: Chat",
    "Topic :: Communications :: Chat :: ICQ",
    "Topic :: Communications :: Chat :: Internet Relay Chat",
    "Topic :: Communications :: Chat :: Unix Talk",
    "Topic :: Communications :: Conferencing",
    "Topic :: Communications :: Email",
    "Topic :: Communications :: Email :: Address Book",
    "Topic :: Communications :: Email :: Email Clients (MUA)",
    "Topic :: Communications :: Email :: Filters",
    "Topic :: Communications :: Email :: Mail Transport Agents",
    "Topic :: Communications :: Email :: Mailing List Servers",
    "Topic :: Communications :: Email :: Post-Office",
    "Topic :: Communications :: Email :: Post-Office :: IMAP",
    "Topic :: Communications :: Email :: Post-Office :: POP3",
    "Topic :: Communications :: FIDO",
    "Topic :: Communications :: Fax",
    "Topic :: Communications :: File Sharing",
    "Topic :: Communications :: File Sharing :: Gnutella",
    "Topic :: Communications :: File Sharing :: Napster",
    "Topic :: Communications :: Ham Radio",
    "Topic :: Communications :: Internet Phone",
    "Topic :: Communications :: Telephony",
    "Topic :: Communications :: Usenet News",
    "Topic :: Database",
    "Topic :: Database :: Database Engines/Servers",
    "Topic :: Database :: Front-Ends",
    "Topic :: Desktop Environment",
    "Topic :: Desktop Environment :: File Managers",
    "Topic :: Desktop Environment :: GNUstep",
    "Topic :: Desktop Environment :: Gnome",
    "Topic :: Desktop Environment :: K Desktop Environment (KDE)",
    "Topic :: Desktop Environment :: K Desktop Environment (KDE) :: Themes",
    "Topic :: Desktop Environment :: PicoGUI",
    "Topic :: Desktop Environment :: PicoGUI :: Applications",
    "Topic :: Desktop Environment :: PicoGUI :: Themes",
    "Topic :: Desktop Environment :: Screen Savers",
    "Topic :: Desktop Environment :: Window Managers",
    "Topic :: Desktop Environment :: Window Managers :: Afterstep",
    "Topic :: Desktop Environment :: Window Managers :: Afterstep :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: Applets",
    "Topic :: Desktop Environment :: Window Managers :: Blackbox",
    "Topic :: Desktop Environment :: Window Managers :: Blackbox :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: CTWM",
    "Topic :: Desktop Environment :: Window Managers :: CTWM :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: Enlightenment",
    "Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Epplets",
    "Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR15",
    "Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR16",
    "Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR17",
    "Topic :: Desktop Environment :: Window Managers :: FVWM",
    "Topic :: Desktop Environment :: Window Managers :: FVWM :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: Fluxbox",
    "Topic :: Desktop Environment :: Window Managers :: Fluxbox :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: IceWM",
    "Topic :: Desktop Environment :: Window Managers :: IceWM :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: MetaCity",
    "Topic :: Desktop Environment :: Window Managers :: MetaCity :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: Oroborus",
    "Topic :: Desktop Environment :: Window Managers :: Oroborus :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: Sawfish",
    "Topic :: Desktop Environment :: Window Managers :: Sawfish :: Themes 0.30",
    "Topic :: Desktop Environment :: Window Managers :: Sawfish :: Themes pre-0.30",
    "Topic :: Desktop Environment :: Window Managers :: Waimea",
    "Topic :: Desktop Environment :: Window Managers :: Waimea :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: Window Maker",
    "Topic :: Desktop Environment :: Window Managers :: Window Maker :: Applets",
    "Topic :: Desktop Environment :: Window Managers :: Window Maker :: Themes",
    "Topic :: Desktop Environment :: Window Managers :: XFCE",
    "Topic :: Desktop Environment :: Window Managers :: XFCE :: Themes",
    "Topic :: Documentation",
    "Topic :: Documentation :: Sphinx",
    "Topic :: Education",
    "Topic :: Education :: Computer Aided Instruction (CAI)",
    "Topic :: Education :: Testing",
    "Topic :: File Formats",
    "Topic :: File Formats :: JSON",
    "Topic :: File Formats :: JSON :: JSON Schema",
    "Topic :: Games/Entertainment",
    "Topic :: Games/Entertainment :: Arcade",
    "Topic :: Games/Entertainment :: Board Games",
    "Topic :: Games/Entertainment :: First Person Shooters",
    "Topic :: Games/Entertainment :: Fortune Cookies",
    "Topic :: Games/Entertainment :: Multi-User Dungeons (MUD)",
    "Topic :: Games/Entertainment :: Puzzle Games",
    "Topic :: Games/Entertainment :: Real Time Strategy",
    "Topic :: Games/Entertainment :: Role-Playing",
    "Topic :: Games/Entertainment :: Side-Scrolling/Arcade Games",
    "Topic :: Games/Entertainment :: Simulation",
    "Topic :: Games/Entertainment :: Turn Based Strategy",
    "Topic :: Home Automation",
    "Topic :: Internet",
    "Topic :: Internet :: File Transfer Protocol (FTP)",
    "Topic :: Internet :: Finger",
    "Topic :: Internet :: Log Analysis",
    "Topic :: Internet :: Name Service (DNS)",
    "Topic :: Internet :: Proxy Servers",
    "Topic :: Internet :: WAP",
    "Topic :: Internet :: WWW/HTTP",
    "Topic :: Internet :: WWW/HTTP :: Browsers",
    "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries",
    "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System",
    "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Message Boards",
    "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: News/Diary",
    "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Page Counters",
    "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Wiki",
    "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
    "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
    "Topic :: Internet :: WWW/HTTP :: Session",
    "Topic :: Internet :: WWW/HTTP :: Site Management",
    "Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking",
    "Topic :: Internet :: WWW/HTTP :: WSGI",
    "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
    "Topic :: Internet :: WWW/HTTP :: WSGI :: Server",
    "Topic :: Internet :: XMPP",
    "Topic :: Internet :: Z39.50",
    "Topic :: Multimedia",
    "Topic :: Multimedia :: Graphics",
    "Topic :: Multimedia :: Graphics :: 3D Modeling",
    "Topic :: Multimedia :: Graphics :: 3D Rendering",
    "Topic :: Multimedia :: Graphics :: Capture",
    "Topic :: Multimedia :: Graphics :: Capture :: Digital Camera",
    "Topic :: Multimedia :: Graphics :: Capture :: Scanners",
    "Topic :: Multimedia :: Graphics :: Capture :: Screen Capture",
    "Topic :: Multimedia :: Graphics :: Editors",
    "Topic :: Multimedia :: Graphics :: Editors :: Raster-Based",
    "Topic :: Multimedia :: Graphics :: Editors :: Vector-Based",
    "Topic :: Multimedia :: Graphics :: Graphics Conversion",
    "Topic :: Multimedia :: Graphics :: Presentation",
    "Topic :: Multimedia :: Graphics :: Viewers",
    "Topic :: Multimedia :: Sound/Audio",
    "Topic :: Multimedia :: Sound/Audio :: Analysis",
    "Topic :: Multimedia :: Sound/Audio :: CD Audio",
    "Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Playing",
    "Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Ripping",
    "Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Writing",
    "Topic :: Multimedia :: Sound/Audio :: Capture/Recording",
    "Topic :: Multimedia :: Sound/Audio :: Conversion",
    "Topic :: Multimedia :: Sound/Audio :: Editors",
    "Topic :: Multimedia :: Sound/Audio :: MIDI",
    "Topic :: Multimedia :: Sound/Audio :: Mixers",
    "Topic :: Multimedia :: Sound/Audio :: Players",
    "Topic :: Multimedia :: Sound/Audio :: Players :: MP3",
    "Topic :: Multimedia :: Sound/Audio :: Sound Synthesis",
    "Topic :: Multimedia :: Sound/Audio :: Speech",
    "Topic :: Multimedia :: Video",
    "Topic :: Multimedia :: Video :: Capture",
    "Topic :: Multimedia :: Video :: Conversion",
    "Topic :: Multimedia :: Video :: Display",
    "Topic :: Multimedia :: Video :: Non-Linear Editor",
    "Topic :: Office/Business",
    "Topic :: Office/Business :: Financial",
    "Topic :: Office/Business :: Financial :: Accounting",
    "Topic :: Office/Business :: Financial :: Investment",
    "Topic :: Office/Business :: Financial :: Point-Of-Sale",
    "Topic :: Office/Business :: Financial :: Spreadsheet",
    "Topic :: Office/Business :: Groupware",
    "Topic :: Office/Business :: News/Diary",
    "Topic :: Office/Business :: Office Suites",
    "Topic :: Office/Business :: Scheduling",
    "Topic :: Other/Nonlisted Topic",
    "Topic :: Printing",
    "Topic :: Religion",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Scientific/Engineering :: Artificial Life",
    "Topic :: Scientific/Engineering :: Astronomy",
    "Topic :: Scientific/Engineering :: Atmospheric Science",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Scientific/Engineering :: Chemistry",
    "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
    "Topic :: Scientific/Engineering :: GIS",
    "Topic :: Scientific/Engineering :: Human Machine Interfaces",
    "Topic :: Scientific/Engineering :: Hydrology",
    "Topic :: Scientific/Engineering :: Image Processing",
    "Topic :: Scientific/Engineering :: Image Recognition",
    "Topic :: Scientific/Engineering :: Information Analysis",
    "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Topic :: Scientific/Engineering :: Medical Science Apps.",
    "Topic :: Scientific/Engineering :: Oceanography",
    "Topic :: Scientific/Engineering :: Physics",
    "Topic :: Scientific/Engineering :: Quantum Computing",
    "Topic :: Scientific/Engineering :: Visualization",
    "Topic :: Security",
    "Topic :: Security :: Cryptography",
    "Topic :: Sociology",
    "Topic :: Sociology :: Genealogy",
    "Topic :: Sociology :: History",
    "Topic :: Software Development",
    "Topic :: Software Development :: Assemblers",
    "Topic :: Software Development :: Bug Tracking",
    "Topic :: Software Development :: Build Tools",
    "Topic :: Software Development :: Code Generators",
    "Topic :: Software Development :: Compilers",
    "Topic :: Software Development :: Debuggers",
    "Topic :: Software Development :: Disassemblers",
    "Topic :: Software Development :: Documentation",
    "Topic :: Software Development :: Embedded Systems",
    "Topic :: Software Development :: Embedded Systems :: Controller Area Network (CAN)",
    "Topic :: Software Development :: Embedded Systems :: Controller Area Network (CAN) :: CANopen",
    "Topic :: Software Development :: Embedded Systems :: Controller Area Network (CAN) :: J1939",
    "Topic :: Software Development :: Internationalization",
    "Topic :: Software Development :: Interpreters",
    "Topic :: Software Development :: Libraries",
    "Topic :: Software Development :: Libraries :: Application Frameworks",
    "Topic :: Software Development :: Libraries :: Java Libraries",
    "Topic :: Software Development :: Libraries :: PHP Classes",
    "Topic :: Software Development :: Libraries :: Perl Modules",
    "Topic :: Software Development :: Libraries :: Pike Modules",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Software Development :: Libraries :: Ruby Modules",
    "Topic :: Software Development :: Libraries :: Tcl Extensions",
    "Topic :: Software Development :: Libraries :: pygame",
    "Topic :: Software Development :: Localization",
    "Topic :: Software Development :: Object Brokering",
    "Topic :: Software Development :: Object Brokering :: CORBA",
    "Topic :: Software Development :: Pre-processors",
    "Topic :: Software Development :: Quality Assurance",
    "Topic :: Software Development :: Testing",
    "Topic :: Software Development :: Testing :: Acceptance",
    "Topic :: Software Development :: Testing :: BDD",
    "Topic :: Software Development :: Testing :: Mocking",
    "Topic :: Software Development :: Testing :: Traffic Generation",
    "Topic :: Software Development :: Testing :: Unit",
    "Topic :: Software Development :: User Interfaces",
    "Topic :: Software Development :: Version Control",
    "Topic :: Software Development :: Version Control :: Bazaar",
    "Topic :: Software Development :: Version Control :: CVS",
    "Topic :: Software Development :: Version Control :: Git",
    "Topic :: Software Development :: Version Control :: Mercurial",
    "Topic :: Software Development :: Version Control :: RCS",
    "Topic :: Software Development :: Version Control :: SCCS",
    "Topic :: Software Development :: Widget Sets",
    "Topic :: System",
    "Topic :: System :: Archiving",
    "Topic :: System :: Archiving :: Backup",
    "Topic :: System :: Archiving :: Compression",
    "Topic :: System :: Archiving :: Mirroring",
    "Topic :: System :: Archiving :: Packaging",
    "Topic :: System :: Benchmark",
    "Topic :: System :: Boot",
    "Topic :: System :: Boot :: Init",
    "Topic :: System :: Clustering",
    "Topic :: System :: Console Fonts",
    "Topic :: System :: Distributed Computing",
    "Topic :: System :: Emulators",
    "Topic :: System :: Filesystems",
    "Topic :: System :: Hardware",
    "Topic :: System :: Hardware :: Hardware Drivers",
    "Topic :: System :: Hardware :: Mainframes",
    "Topic :: System :: Hardware :: Symmetric Multi-processing",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB)",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Audio",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Audio/Video (AV)",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Communications Device Class (CDC)",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Diagnostic Device",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Hub",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Human Interface Device (HID)",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Mass Storage",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Miscellaneous",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Printer",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Smart Card",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Vendor",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Video (UVC)",
    "Topic :: System :: Hardware :: Universal Serial Bus (USB) :: Wireless Controller",
    "Topic :: System :: Installation/Setup",
    "Topic :: System :: Logging",
    "Topic :: System :: Monitoring",
    "Topic :: System :: Networking",
    "Topic :: System :: Networking :: Firewalls",
    "Topic :: System :: Networking :: Monitoring",
    "Topic :: System :: Networking :: Monitoring :: Hardware Watchdog",
    "Topic :: System :: Networking :: Time Synchronization",
    "Topic :: System :: Operating System",
    "Topic :: System :: Operating System Kernels",
    "Topic :: System :: Operating System Kernels :: BSD",
    "Topic :: System :: Operating System Kernels :: GNU Hurd",
    "Topic :: System :: Operating System Kernels :: Linux",
    "Topic :: System :: Power (UPS)",
    "Topic :: System :: Recovery Tools",
    "Topic :: System :: Shells",
    "Topic :: System :: Software Distribution",
    "Topic :: System :: System Shells",
    "Topic :: System :: Systems Administration",
    "Topic :: System :: Systems Administration :: Authentication/Directory",
    "Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP",
    "Topic :: System :: Systems Administration :: Authentication/Directory :: NIS",
    "Topic :: Terminals",
    "Topic :: Terminals :: Serial",
    "Topic :: Terminals :: Telnet",
    "Topic :: Terminals :: Terminal Emulators/X Terminals",
    "Topic :: Text Editors",
    "Topic :: Text Editors :: Documentation",
    "Topic :: Text Editors :: Emacs",
    "Topic :: Text Editors :: Integrated Development Environments (IDE)",
    "Topic :: Text Editors :: Text Processing",
    "Topic :: Text Editors :: Word Processors",
    "Topic :: Text Processing",
    "Topic :: Text Processing :: Filters",
    "Topic :: Text Processing :: Fonts",
    "Topic :: Text Processing :: General",
    "Topic :: Text Processing :: Indexing",
    "Topic :: Text Processing :: Linguistic",
    "Topic :: Text Processing :: Markup",
    "Topic :: Text Processing :: Markup :: HTML",
    "Topic :: Text Processing :: Markup :: LaTeX",
    "Topic :: Text Processing :: Markup :: Markdown",
    "Topic :: Text Processing :: Markup :: SGML",
    "Topic :: Text Processing :: Markup :: VRML",
    "Topic :: Text Processing :: Markup :: XML",
    "Topic :: Text Processing :: Markup :: reStructuredText",
    "Topic :: Utilities",
    "Typing :: Stubs Only",
    "Typing :: Typed",
]

classifiers: Set[str] = set(sorted_classifiers)


# A mapping from the deprecated classifier name to a list of zero or more valid
# classifiers that should replace it
deprecated_classifiers: Dict[str, List[str]] = {
    "Framework :: Django CMS :: 4.2": ["Framework :: Django CMS :: 5.0"],
    "License :: OSI Approved :: Intel Open Source License": [],
    "License :: OSI Approved :: Jabber Open Source License": [],
    "License :: OSI Approved :: MITRE Collaborative Virtual Workspace License (CVW)": [],
    "License :: OSI Approved :: Sun Industry Standards Source License (SISSL)": [],
    "License :: OSI Approved :: X.Net License": [],
    "Natural Language :: Ukranian": ["Natural Language :: Ukrainian"],
    "Topic :: Communications :: Chat :: AOL Instant Messenger": [],
}


# All classifiers, including deprecated classifiers
all_classifiers: List[str] = sorted(
    sorted_classifiers + list(deprecated_classifiers.keys())
)

__all__ = [
    "all_classifiers",
    "classifiers",
    "deprecated_classifiers",
    "sorted_classifiers",
]



import pandas as pd
from typing import Literal, Optional, Union
from scipy.signal import savgol_filter

from neptoon.logging import get_logger
from neptoon.columns import ColumnInfo
from neptoon.utils import parse_resolution_to_timedelta

core_logger = get_logger()


class SmoothData:
    """
    A class for smoothing data using a variety of different methods.

    This class provides functionality to smooth time series data using
    different methods such as rolling mean and Savitzky-Golay filter. It
    supports both count-based and time-based windows for the rolling
    mean method.

    """

    def __init__(
        self,
        data: pd.Series,
        column_to_smooth: str,
        smooth_method: Literal[
            "rolling_mean", "savitsky_golay"
        ] = "rolling_mean",
        window: str = "12h",
        min_proportion_good_data: float = 0.5,
        poly_order: Optional[int] = None,
        auto_update_final_col: bool = True,
    ):
        """
        Attributes for SmoothData

        Parameters
        ----------
        data : pd.Series
            Data to be smoothed - must be time indexed
        column_to_smooth : str
            The column name of the column to be smoothed.
        smooth_method : Literal["rolling_mean", "savitsky_golay"],
            The smooth method to apply, by default "rolling_mean"
        window : str , optional
            The window size for smoothing. Time based str value e.g.,
            30m, 1h, 6h, 1d by default 24h. Converted to timedelta for
            use in smoothing.
        min_proportion_good_data : float, optional
            The minimum proportion of available data in the window for a
            average to be taken, provided as a decimal.
        poly_order : Optional[int], optional
            Poly order to apply (savitsky golay only), by default None
        auto_update_final_col : bool, optional
            Whether to update the ColumnInfo object to represent the new
            column as _FINAL, by default True
        """
        self.data = data
        self.column_to_smooth = column_to_smooth
        self.smooth_method = smooth_method
        self.window = window
        self.min_proportion_good_data = min_proportion_good_data
        self.poly_order = poly_order
        self.auto_update_final_col = auto_update_final_col

        # Placeholders
        self.window_as_timedelta = None

        self._validate_inputs()

    def _validate_inputs(self):
        """
        Validate the inputs in SmoothData class to ensure selected
        smoothing method can be applied with given data.

        Raises
        ------
        ValueError
            If not DateTmeIndex in data
        ValueError
            If unsupported method is supplied
        ValueError
            If invalid time string supplied for rolling mean
        """
        if not isinstance(self.data.index, pd.DatetimeIndex):
            message = "Data index must be a DatetimeIndex"
            core_logger.error(message)
            raise ValueError(message)
        if not isinstance(self.column_to_smooth, str):
            message = "column_to_smooth must be a string type"
            core_logger.error(message)
            raise ValueError(message)
        if self.smooth_method not in ["rolling_mean", "savitsky_golay"]:
            message = (
                "smooth_method must be either 'rolling_mean' or "
                "'savitsky_golay'"
            )
            core_logger.error(message)
            raise ValueError(message)
        if self.smooth_method == "savitsky_golay":
            self._switch_to_rolling()  # TODO : change this when implementing savitsky_golay
            # self._validate_savitsky_golay_params()
        if self.smooth_method == "rolling_mean":
            self._validate_rolling_mean_params()
        if isinstance(self.window, str):
            self.window_as_timedelta = parse_resolution_to_timedelta(
                resolution_str=self.window
            )
        else:
            message = (
                f"Invalid time string for window: {self.window}"
                "`window` must be a string type denoting the time window for "
                "smoothing. E.g., 30 minutes = `30m`, 2 hours = `2h`, 1 day = `1d`"
            )
            core_logger.error(message)
            raise ValueError(message)

    def _validate_rolling_mean_params(self):
        if not isinstance(self.window, str):
            message = (
                "window must be given as a string (e.g., 1h, 1d, 30m). "
                f"Window was given as {self.window}"
            )
            core_logger.error(message)
            raise ValueError(message)

    def _switch_to_rolling(self):
        self.smooth_method = "rolling_mean"
        message = (
            "Savitsky-Golay smoothing is coming in a future update."
            "For this processing pipeline smoothing has been changed to rolling mean."
        )
        core_logger.info(message)
        print(message)

    def _validate_savitsky_golay_params(self):
        """
        Validates that the parameters are appropriate for using savitsky
        golay smoothing.

        Raises
        ------
        ValueError
            error if polyorder not supplied
        ValueError
            error if time based window given
        """
        if self.poly_order is None:
            message = (
                "poly_order cannot be None when implementing SG filter. "
                "Either change the smoothing method or supply a poly_order value"
            )
            core_logger.error(message)
            raise ValueError(message)
        if isinstance(self.window, str):
            message = "Time-based window is not supported for Savitzky-Golay smoothing"
            core_logger.error(message)
            raise ValueError(message)

    def _update_column_name_config(
        self,
        possible_names=[
            str(ColumnInfo.Name.SOIL_MOISTURE_VOL),
            str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL),
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH),
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL),
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT),
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL),
        ],
    ):
        """
        Updates the value for the '_FINAL' value in ColumnInfo.
        Currently restricted to EPI_NEUTRONS and SOIL_MOISTURE. These
        are most likely to be smoothed and used throughout the code.

        The possible_names parameter is included to give flexibility in
        future with updating the names. (i.e., we could include support
        for supplying other columns to be automatically updated with
        '_FINAL' after smoothing)

        Parameters
        ----------
        possible_names : list, optional
            _description_, by default [
            str(ColumnInfo.Name.SOIL_MOISTURE),
            str(ColumnInfo.Name.SOIL_MOISTURE_FINAL),
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH),
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL),
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT),
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL), ]
        """
        if not self.auto_update_final_col:
            return
        if self.column_to_smooth not in possible_names:
            message = (
                f"{self.column_to_smooth} is not supported for automatic "
                "updating of ColumnInfo. Please turn off auto_update_final_col"
            )
            core_logger.error(message)
            raise ValueError(message)

        new_column_name = self.create_new_column_name()
        if self.column_to_smooth in [
            str(ColumnInfo.Name.SOIL_MOISTURE_VOL),
            str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL),
        ]:
            ColumnInfo.relabel(
                ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL,
                new_label=new_column_name,
            )
        elif self.column_to_smooth in [
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH),
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL),
        ]:
            ColumnInfo.relabel(
                ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL,
                new_label=new_column_name,
            )
        elif self.column_to_smooth in [
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT),
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL),
        ]:
            ColumnInfo.relabel(
                ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL,
                new_label=new_column_name,
            )

    def _get_min_obs_for_rolling_mean(
        self,
        data_to_smooth: pd.Series,
        min_proportion_good_data: float = 0.5,
    ):
        """
        Calculates the number of observations that equates to half the
        timedelta window.

        Necessary as pd.Series.rolling(min_periods) only accepts
        integers (not timedelta)

        Parameters
        ----------
        data_to_smooth : pd.Series
            data on which smoothing will be done (datetime index)

        Returns
        -------
        int
            The number of observatinos that make up the half the time
            delta
        """
        freq = data_to_smooth.index.to_series().diff().median()
        min_obs = int(
            (self.window_as_timedelta * min_proportion_good_data) / freq
        )
        return min_obs

    def _apply_rolling_mean(self, data_to_smooth: pd.Series):
        """
        Applies a rolling mean smoothing

        Parameters
        ----------
        data_to_smooth : pd.Series
            pd.Series of data to smooth

        Returns
        -------
        pd.Series
            The smoothed data
        """
        min_obs = self._get_min_obs_for_rolling_mean(
            data_to_smooth=data_to_smooth
        )

        smoothed = data_to_smooth.rolling(
            window=self.window_as_timedelta,
            min_periods=min_obs,
            center=False,
        ).mean()

        return smoothed.round()

    def _apply_savitsky_golay(self, data_to_smooth):
        """
        Applies the savitsky golay smoothing technique

        Parameters
        ----------
        data_to_smooth : str
            column name to be smoothed

        Returns
        -------
        pd.Series
            smoothed series

        #TODO
        THIS DOESN'T WORK WHEN NAN VALUES ARE IN WINDOW.

        """

        smoothed = savgol_filter(
            x=data_to_smooth,
            window_length=self.window,
            polyorder=self.poly_order,
        )
        return pd.Series(smoothed, index=self.data.index).round()

    def create_new_column_name(self):
        """
        Creates a new column name based on the supplied column_to_smooth
        name. This depends on method and parameters used.

        Returns
        -------
        str
            New column name
        """
        og_column_name = self.column_to_smooth + "_"

        if self.smooth_method == "rolling_mean":
            add_on = f"rollingmean_{str(self.window)}"
        elif self.smooth_method == "savitsky_golay":
            add_on = f"savgol_{str(self.window)}_{str(self.poly_order)}"

        return og_column_name + add_on

    def apply_smoothing(self):
        """
        Function to apply smoothing to a Series of data. It is presumed
        that the appropriate attributes have been assigned already.

        Returns
        -------
        pd.Series
            The smoothed Series
        """
        if self.smooth_method == "rolling_mean":
            self._update_column_name_config()
            return self._apply_rolling_mean(self.data)
        elif self.smooth_method == "savitsky_golay":
            self._update_column_name_config()
            return self._apply_savitsky_golay(self.data)



"""
Functions related neutron corrections for changes to above ground
biomass.

Functions in this module:

    above_ground_biomass_correction_baatz2015
"""

from neptoon.logging import get_logger

core_logger = get_logger()


def above_ground_biomass_correction_baatz2015(above_ground_biomass):
    """
    Provides the correction value to account for changes in above ground
    biomass.

    Parameters
    ----------
    above_ground_biomass : float
        above ground biomass in kilograms per meter squared (kg/m^2)

    Returns
    -------
    c_factor: float
        Correction factor to be multiplied with neutron counts.
    """
    c_factor = 1 / (1 - (0.009 * above_ground_biomass))
    return c_factor


def above_ground_biomass_correction_morris2024(biomass_water_equivalent):
    """
    Provides the correction value to account for changes in above ground
    biomass.

    Parameters
    ----------
    biomass_water_equivalent : float
        biomass_water_equivalent in mm (kg/m^2). To convert above ground
        biomass to biomass water equivalent you can multiply by the
        water equivelant of cellulose (0.494)

    Returns
    -------
    c_factor: float
        Correction factor to be multiplied with neutron counts.
    """
    c_factor = 1 / (1 - (0.01 * biomass_water_equivalent))
    return c_factor



import argparse

from hatchling.cli.build import build_command
from hatchling.cli.dep import dep_command
from hatchling.cli.metadata import metadata_command
from hatchling.cli.version import version_command


def hatchling() -> int:
    parser = argparse.ArgumentParser(prog='hatchling', allow_abbrev=False)
    subparsers = parser.add_subparsers()

    defaults = {'metavar': ''}

    build_command(subparsers, defaults)
    dep_command(subparsers, defaults)
    metadata_command(subparsers, defaults)
    version_command(subparsers, defaults)

    kwargs = vars(parser.parse_args())
    try:
        command = kwargs.pop('func')
    except KeyError:
        parser.print_help()
    else:
        command(**kwargs)

    return 0



import pytest
from unittest.mock import patch


@pytest.fixture(autouse=True)
def mock_logging(request):
    """
    Turn of logging for tests.
    """
    if "test_logging" in request.keywords:
        yield
    else:
        with patch("logging.Logger") as MockLogger:
            yield MockLogger



from enum import Enum
from dataclasses import dataclass
from typing import Dict, Optional, Any, Set, Type
from neptoon.columns import ColumnInfo


class QAMethod(Enum):
    """
    The methods that can be selected in neptoon. Methods are implemented
    using SaQC.

    For methods that use the same underlying SaQC function but with different
    configurations (like ABOVE_N0 and BELOW_N0_FACTOR both using flagGeneric),
    we use a tuple to store both the SaQC method and a discriminator.
    """

    RANGE_CHECK = ("flagRange", None)
    SPIKE_UNILOF = ("flagUniLOF", None)
    CONSTANT = ("flagConstants", None)
    ABOVE_N0 = ("flagRange", "above_n0")
    BELOW_N0_FACTOR = ("flagRange", "below_n0")

    @property
    def saqc_method(self) -> str:
        """Returns the underlying SaQC method name."""
        return self.value[0]

    @property
    def variant(self) -> str:
        """Returns the variant discriminator if any."""
        return self.value[1]


class QATarget(Enum):
    """
    The target data for the quality assessment selection.
    """

    RAW_EPI_NEUTRONS = str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL)
    CORRECTED_EPI_NEUTRONS = str(
        ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL
    )
    RELATIVE_HUMIDITY = str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY)
    AIR_PRESSURE = str(ColumnInfo.Name.AIR_PRESSURE)
    TEMPERATURE = str(ColumnInfo.Name.AIR_TEMPERATURE)
    SOIL_MOISTURE_VOL = str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL)
    CUSTOM = "custom"


@dataclass(frozen=True)
class ParameterSpec:
    """
    Specification for a single parameter.

    Attributes
    ----------
    name : str
        Parameter name
    description : str
        Parameter description
    units : Optional[str]
        Parameter units (if applicable)
    default : Any
        Default value (if optional)
    """

    name: str
    description: str
    optional: bool = False
    units: Optional[str] = None
    default: Any = None
    saqc_name: str = None


class MethodParameters:
    """
    Base class for method parameter specifications.
    Subclasses define parameter requirements for each method.
    """

    saqc_web: str = None
    essential_params: Set[ParameterSpec] = set()
    optional_params: Set[ParameterSpec] = set()


class AboveN0Parameters(MethodParameters):

    saqc_web = "https://rdm-software.pages.ufz.de/saqc/_api/saqc.SaQC.html#saqc.SaQC.flagGeneric"

    essential_params: Set[ParameterSpec] = {
        ParameterSpec(
            name="N0",
            description="The N0 calibration number",
            units="neutron counts per hour",
            saqc_name="N0",
        ),
        ParameterSpec(
            name="percent_maximum",
            description="Fraction above N0 to flag. Commonly set to 1.075",
            units="decimal",
            saqc_name="percent_maximum",
        ),
    }


class BelowFactorofN0Parameters(MethodParameters):
    """Parameter specifications for below N0 factor check method."""

    saqc_web = "https://rdm-software.pages.ufz.de/saqc/_api/saqc.SaQC.html#saqc.SaQC.flagGeneric"

    essential_params: Set[ParameterSpec] = {
        ParameterSpec(
            name="N0",
            description="The N0 calibration number",
            units="neutron counts per hour",
            saqc_name="N0",
        ),
        ParameterSpec(
            name="percent_minimum",
            description="The fraction of N0 below which to flag",
            units="decimal",
            saqc_name="percent_minimum",
        ),
    }


class RangeCheckParameters(MethodParameters):
    """Parameter specifications for range check method."""

    saqc_web = "https://rdm-software.pages.ufz.de/saqc/_api/saqc.SaQC.html#saqc.SaQC.flagRange"

    essential_params = {
        ParameterSpec(
            name="min",
            description="Minimum acceptable value",
            units="data units",
            saqc_name="min",
        ),
        ParameterSpec(
            name="max",
            description="Maximum acceptable value",
            units="data units",
            saqc_name="max",
        ),
    }


class UniLOFParameters(MethodParameters):
    """Parameter specifications for range check method."""

    saqc_web = "https://rdm-software.pages.ufz.de/saqc/_api/saqc.SaQC.html#saqc.SaQC.flagUniLOF"

    essential_params = {}

    optional_params = {
        ParameterSpec(
            name="periods_in_calculation",
            description=str(
                "Number of periods to be included into the LOF calculation"
            ),
            units="time steps",
            default="20",
            saqc_name="n",
        ),
        ParameterSpec(
            name="threshold",
            description="Threshold for flagging",
            units="decimal",
            default="1.5",
            saqc_name="thresh",
        ),
        ParameterSpec(
            name="algorithm",
            description=(
                "Algorithm used for calculating the n-nearest "
                "neighbors needed for LOF calculation.\n"
                "    ['ball_tree', 'kd_tree', 'brute', 'auto']"
            ),
            units="Literal",
            default="ball_tree",
            saqc_name="algorithm",
        ),
    }


class WhatParamsDoINeed:
    """
    Helper class for discovering parameter requirements for QA methods.

    Parameters
    ----------
    method : SaQCMethodMap
        The quality assessment method to investigate

    Example
    -------
    >>> from neptoon.quality_control import WhatParamsDoINeed, QAMethod
    >>>
    >>> WhatParamsDoINeed(QAMethod.RANGE_CHECK)

    """

    def __init__(self, method: QAMethod):
        self.method = method
        self._param_class = ParameterRegistry.get_parameter_class(method)
        self.show_all_params()

    def show_required_params(self):
        """
        Display essential parameters for the method.
        """
        print(f"\nRequired parameters for {self.method}:")
        print("-" * 50)
        for param in self._param_class.essential_params:
            units_str = f"[{param.units}]" if param.units else ""
            print(f"{param.name} - {units_str}:")
            print(f"    {param.description}")

    def show_optional_params(self):
        """
        Display optional parameters for the method.
        """
        print(f"\nOptional parameters for {self.method}:")
        print("-" * 50)
        for param in self._param_class.optional_params:
            units_str = f"[{param.units}]" if param.units else ""
            default_str = f" (default: {param.default})"
            print(f"{param.name} - {units_str}{default_str}:")
            print(f"    {param.description}")

    def show_link_to_site(self):
        """
        Adds a link to the SaQC documentation
        """
        print(f"\nFurther information about {self.method}:")
        print("-" * 50)
        print(self._param_class.saqc_web)

    def show_all_params(self):
        """
        Display the params for the method.
        """
        self.show_required_params()
        self.show_optional_params()
        self.show_link_to_site()


class ParameterRegistry:
    """
    Central registry for mapping quality assessment methods to their
    parameter types.

    This class manages the relationships between methods and their
    parameter specifications, providing type safety and validation.
    """

    _registry: Dict[QAMethod, Type[MethodParameters]] = {
        QAMethod.RANGE_CHECK: RangeCheckParameters,
        QAMethod.ABOVE_N0: AboveN0Parameters,
        QAMethod.BELOW_N0_FACTOR: BelowFactorofN0Parameters,
        QAMethod.SPIKE_UNILOF: UniLOFParameters,
    }

    @classmethod
    def get_parameter_class(cls, method: QAMethod):
        """
        Get the parameter class for a given method.

        Parameters
        ----------
        method : SaQCMethodMap
            The quality assessment method

        Returns
        -------
        Type[BaseParameters]
            The corresponding parameter class

        Raises
        ------
        KeyError
            If the method is not registered
        """
        if method not in cls._registry:
            raise KeyError(
                f"No parameter specification found for method {method}"
            )
        return cls._registry[method]


class QAConfigRegistry:
    """
    Registry system for mapping strings as found in the configuration
    file to specific QATarget and QAMethods.
    """

    _target_mapping = {
        "raw_neutrons": QATarget.RAW_EPI_NEUTRONS,
        "corrected_neutrons": QATarget.CORRECTED_EPI_NEUTRONS,
        "air_relative_humidity": QATarget.RELATIVE_HUMIDITY,
        "air_pressure": QATarget.AIR_PRESSURE,
        "temperature": QATarget.TEMPERATURE,
        "soil_moisture": QATarget.SOIL_MOISTURE_VOL,
    }

    _method_mapping = {
        "flag_range": QAMethod.RANGE_CHECK,
        "spike_uni_lof": QAMethod.SPIKE_UNILOF,
        "constant": QAMethod.CONSTANT,
        "greater_than_N0": QAMethod.ABOVE_N0,
        "below_N0_factor": QAMethod.BELOW_N0_FACTOR,
    }

    @classmethod
    def get_target(cls, target: str) -> QATarget:
        return cls._target_mapping[target]

    @classmethod
    def get_method(cls, method: str) -> QAMethod:
        return cls._method_mapping[method]



import pandas as pd
import requests
import pandas.testing as pdt
from pathlib import Path
from neptoon.external.nmdb_data_collection import (
    DateTimeHandler,
    NMDBConfig,
    NMDBDataHandler,
    DataFetcher,
    CacheHandler,
    DataManager,
)

"""
DateTimeHandler Tests
"""


def test_standardize_date_input():
    """Tests to ensure teh standardize date works"""
    assert DateTimeHandler.standardize_date_input("2024-01-23") == "2024-01-23"
    assert DateTimeHandler.standardize_date_input("23/01/2024") == "2024-01-23"
    assert DateTimeHandler.standardize_date_input("01-23-2024") == "2024-01-23"
    assert (
        DateTimeHandler.standardize_date_input(pd.Timestamp("2024-01-23"))
        == "2024-01-23"
    )
    assert DateTimeHandler.standardize_date_input(pd.to_datetime("2024-01-23"))
    standardized_date = DateTimeHandler.standardize_date_input("invalid-date")
    assert (
        standardized_date is None
    ), "The method should return None for invalid date"


"""
CacheHandler Tests
"""


def test_read_cache():
    """Test the read cache and make sure the format is as expected"""
    config = NMDBConfig(
        start_date_wanted="2015-10-10",
        end_date_wanted="2016-10-10",
    )
    cache_handler = CacheHandler(config)
    mock_file_path = (
        Path(__file__).parent / "mock_data" / "example_cache_1516.csv"
    )
    # monkeypatch.setattr(config, "cache_exists", True)
    config.cache_exists = True
    cache_handler.cache_file_path = mock_file_path
    df = cache_handler.read_cache()

    assert isinstance(df, pd.DataFrame)
    assert not df.empty
    assert "count" in df.columns
    assert isinstance(df.index, pd.DatetimeIndex)


def test_check_cache_range():
    """Test on logic to collect cache range"""
    config = NMDBConfig(
        start_date_wanted="2015-10-10",
        end_date_wanted="2016-10-10",
    )
    cache_handler = CacheHandler(config)
    mock_file_path = (
        Path(__file__).parent / "mock_data" / "example_cache_1516.csv"
    )
    config.cache_exists = True
    cache_handler.cache_file_path = mock_file_path
    cache_handler.check_cache_range()
    df = pd.read_csv(mock_file_path)
    df["datetime"] = pd.to_datetime(df["datetime"])

    assert config.cache_start_date == df["datetime"].min().date()
    assert config.cache_end_date == df["datetime"].max().date()


"""
DataFetcher Tests
"""


def test_get_ymd_from_date(monkeypatch):
    """Test the ymd parsing step in DataFetcher"""
    monkeypatch.setattr(
        DateTimeHandler, "standardize_date_input", lambda x: "2016-10-28"
    )
    config = NMDBConfig(
        start_date_wanted="2015-10-10",
        end_date_wanted="2016-10-10",
    )
    data_fetcher = DataFetcher(config)
    year, month, day = data_fetcher.get_ymd_from_date("some_date")

    assert year == "2016"
    assert month == "10"
    assert day == "28"


def test_create_nmdb_url_config1(monkeypatch):
    """Test the creation of the NMDB http URL"""

    def mock_get_ymd_from_date(date1, date2):
        return ("2015", "10", "10")

    config = NMDBConfig(
        start_date_wanted="2015-10-10",
        end_date_wanted="2015-10-10",
        station="JUNG",
        nmdb_table="revori",
        resolution="60",
        start_date_needed=None,
        end_date_needed=None,
    )

    monkeypatch.setattr(
        DataFetcher, "get_ymd_from_date", mock_get_ymd_from_date
    )

    data_fetcher = DataFetcher(config)
    expected_url = (
        "http://nest.nmdb.eu/draw_graph.php?wget=1&stations"
        "[]=JUNG&tabchoice=revori&dtype=corr_for_efficiency"
        "&tresolution=60&yunits=0&date_choice=bydate"
        "&start_day=10&start_month=10&start_year=2015&start_hour=0"
        "&start_min=0&end_day=10&end_month=10&end_year=2015"
        "&end_hour=23&end_min=59&output=ascii"
    )
    url = data_fetcher.create_nmdb_url()
    assert url == expected_url


class MockHTTPResponse:
    def __init__(self, text_data):
        self.text = text_data

    def raise_for_status(self):
        pass


def test_fetch_data_http(monkeypatch):
    """Test fetching data from http server with mock http"""
    file_location = (
        Path(__file__).parent / "mock_data" / "mock_http_2015_2016.txt"
    )

    text_to_assert = """#        STATION: JUNG
#     START TIME: 2015-10-10 00:00:00 UTC
#       END TIME: 2016-10-10 23:00:00 UTC
#     NMDB TABLE: revised original
#    REV EQ. ORI: No
#    REV. VALUES: 4543"""

    config = NMDBConfig(
        start_date_wanted="2015-10-10",
        end_date_wanted="2016-10-10",
    )

    def mock_get(*args, **kwargs):
        with open(
            file_location,
            "r",
        ) as file:
            mock_text = file.read()
        return MockHTTPResponse(mock_text)

    monkeypatch.setattr(requests, "get", mock_get)

    handler = NMDBDataHandler(config)

    result = handler.data_fetcher.fetch_data_http()
    assert text_to_assert in result


def test_parse_http_date():
    config = NMDBConfig(
        start_date_wanted="2015-10-10",
        end_date_wanted="2015-10-10",
        station="JUNG",
    )
    mock_file_path = (
        Path(__file__).parent / "mock_data" / "mock_http_2015_2016.txt"
    )
    assert_file_path = (
        Path(__file__).parent / "mock_data" / "example_cache_1516.csv"
    )

    df_to_compare = pd.read_csv(assert_file_path)
    df_to_compare["datetime"] = pd.to_datetime(df_to_compare["datetime"])
    df_to_compare.set_index("datetime", inplace=True)
    df_to_compare.index = df_to_compare.index.tz_localize("UTC")

    data_fetcher = DataFetcher(config)
    with open(mock_file_path, "r") as file:
        raw_data = file.read()

    data = data_fetcher.parse_http_data(raw_data)

    pdt.assert_frame_equal(data, df_to_compare)


"""
DataManager Tests
"""


def test_check_if_need_extra_data(monkeypatch):
    """Test to check logic on whether extra dates are needed"""
    config = NMDBConfig(
        start_date_wanted="2015-10-10",
        end_date_wanted="2016-10-10",
    )
    monkeypatch.setattr(CacheHandler, "check_cache_range", lambda x: None)
    cache_handler = CacheHandler(config)
    data_fetcher = DataFetcher(config)
    data_manager = DataManager(config, cache_handler, data_fetcher)
    config.cache_start_date = pd.to_datetime("2015-10-10").date()
    config.cache_end_date = pd.to_datetime("2016-10-10").date()

    data_manager.check_if_need_extra_data()

    assert data_manager.need_data_before_cache is False
    assert data_manager.need_data_after_cache is False

    config.cache_start_date = pd.to_datetime("2015-11-10").date()
    config.cache_end_date = pd.to_datetime("2016-10-10").date()

    data_manager.check_if_need_extra_data()

    assert data_manager.need_data_before_cache is True
    assert data_manager.need_data_after_cache is False

    config.cache_start_date = pd.to_datetime("2015-11-10").date()
    config.cache_end_date = pd.to_datetime("2016-09-10").date()

    data_manager.check_if_need_extra_data()

    assert data_manager.need_data_before_cache is True
    assert data_manager.need_data_after_cache is True

    config.cache_start_date = pd.to_datetime("2015-11-10").date()
    config.cache_end_date = pd.to_datetime("2016-09-10").date()

    data_manager.check_if_need_extra_data()

    assert data_manager.need_data_before_cache is True
    assert data_manager.need_data_after_cache is True


def test_combine_cache_and_new_data(monkeypatch):
    config = NMDBConfig(
        start_date_wanted="2015-10-10",
        end_date_wanted="2016-10-10",
    )
    cache_handler = CacheHandler(config)
    data_fetcher = DataFetcher(config)
    data_manager = DataManager(config, cache_handler, data_fetcher)

    df_cache_path = (
        Path(__file__).parent / "mock_data" / "example_cache_merge1.csv"
    )
    df_cache = pd.read_csv(df_cache_path)
    df_cache["datetime"] = pd.to_datetime(df_cache["datetime"])
    df_cache.set_index("datetime", inplace=True)

    print(df_cache_path)

    df_download_path = (
        Path(__file__).parent / "mock_data" / "example_cache_merge2.csv"
    )
    df_download = pd.read_csv(df_download_path)
    df_download["datetime"] = pd.to_datetime(df_download["datetime"])
    df_download.set_index("datetime", inplace=True)

    print(df_download_path)

    df_assert_path = (
        Path(__file__).parent / "mock_data" / "example_cache_merge_assert.csv"
    )
    df_assert = pd.read_csv(df_assert_path)
    df_assert["datetime"] = pd.to_datetime(df_assert["datetime"])
    df_assert.set_index("datetime", inplace=True)
    df_assert.index = df_assert.index.tz_localize("UTC")

    combined_df = data_manager.combine_cache_and_new_data(
        df_cache, df_download
    )

    assert isinstance(combined_df, pd.DataFrame)
    assert not combined_df.empty
    assert "count" in combined_df.columns
    pdt.assert_frame_equal(combined_df, df_assert)


# TODO:
# NMDBDataHandler Tests - canary
